#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility


def define_simulation(sim, threads=1):
    um = gate.g4_units.um
    cm = gate.g4_units.cm
    m = gate.g4_units.m
    keV = gate.g4_units.keV

    sim.running_verbose_level = gate.logger.RUN
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_engine = "MersenneTwister"
    sim.random_seed = 123654789
    sim.number_of_threads = threads
    print(sim)

    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    waterbox = sim.add_volume("Box", "Waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    sim.physics_manager.physics_list_name = "QGSP_BERT_EMV"
    global_cut = 700 * um
    sim.physics_manager.global_production_cuts.gamma = global_cut
    sim.physics_manager.global_production_cuts.electron = global_cut
    sim.physics_manager.global_production_cuts.positron = global_cut
    sim.physics_manager.global_production_cuts.proton = global_cut

    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 80 * keV
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 200000 / sim.number_of_threads

    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.track_types_flag = True


def test_output(sim):
    # get output
    paths = utility.get_default_test_paths(
        __file__, "gate_test004_simulation_stats_actor"
    )

    stats = sim.get_actor("Stats")
    print(stats)
    stats_ref = utility.read_stat_file(paths.gate_output / "stat.txt")
    stats_ref.counts.runs = sim.number_of_threads
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.01)

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test085_free_flight_helpers import *
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, None, output_folder="test085_phsp")

    # create the simulation
    sim = gate.Simulation()
    sim.number_of_threads = 4
    # sim.visu = True
    source, actors = create_simulation_test085(
        sim,
        paths,
        simu_name="ref",
        ac=5e4,
        use_spect_head=False,
        use_spect_arf=False,
        use_phsp=True,
    )

    # go
    sim.run()
    stats = sim.get_actor("stats")
    print(stats)

    # compare histo
    is_ok = utility.compare_root3(
        paths.output_ref / "phsp_sphere_ref.root",
        paths.output / "phsp_sphere_ref.root",
        "phsp_sphere",
        "phsp_sphere",
        keys1=None,
        keys2=None,
        tols=[0.01, 0.7, 1.4, 1.4, 0.01],
        scalings1=[1] * 5,
        scalings2=[1] * 5,
        img=paths.output / "test085_phsp_ref.png",
    )

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.pet.philipsvereos as pet_vereos
import opengate.contrib.phantoms.necr as phantom_necr
from opengate.tests import utility
from opengate.userhooks import check_production_cuts
from opengate.sources.utility import get_rad_yield


def create_pet_simulation(sim, paths, debug=False, create_mat=False):
    """
    Simulation of a PET VEREOS with NEMA NECR phantom.
    - phantom is a simple cylinder and linear source
    - output is hits and singles only (no coincidences)
    - also digitizer is simplified: only raw hits and adder (for singles)
    """

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.check_volumes_overlap = False
    sim.random_seed = 123456789
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq
    MBq = Bq * 1e6
    sec = gate.g4_units.second

    #  change world size
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    # add a PET VEREOS
    sim.volume_manager.add_material_database(paths.gate_data / "GateMaterials_pet.db")
    if not debug:
        pet = pet_vereos.add_pet(sim, "pet", create_housing=True, create_mat=create_mat)
    else:
        pet = pet_vereos.add_pet(
            sim, "pet", create_housing=True, create_mat=create_mat, debug=True
        )

    # add table
    bed = pet_vereos.add_table(sim, "pet")

    # add NECR phantom
    phantom = phantom_necr.add_necr_phantom(sim, "phantom")

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.set_production_cut("world", "all", 1 * m)

    reg1 = sim.physics_manager.add_region("reg1")
    reg1.production_cuts.all = 10 * mm
    reg1.associate_volume(phantom)
    reg1.associate_volume(bed)

    reg2 = sim.physics_manager.add_region("reg2")
    reg2.production_cuts.all = 0.1 * mm
    reg2.associate_volume(f"{pet.name}_crystal")

    # default source for tests
    source = phantom_necr.add_necr_source(sim, phantom)
    total_yield = get_rad_yield("F18")
    print("Yield for F18 (nb of e+ per decay) : ", total_yield)
    source.activity = 3000 * Bq * total_yield
    source.activity = 1787.914158 * MBq * total_yield
    source.half_life = 6586.26 * sec
    source.energy.type = "F18_analytic"  # WARNING not ok, but similar to previous Gate
    # source.energy.type = "F18"  # this is the correct F18 e+ source

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "Stats")
    s.track_types_flag = True

    # set user hook function
    sim.user_hook_after_init = check_production_cuts

    for vol in sim.volume_manager.volumes.values():
        if "crystal" in vol.name:
            return vol
    gate.exception.fatal("Could not find any crystal volume.")


def add_digitizer(sim, paths, nb, crystal):
    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    hc.attached_to = crystal.name
    hc.authorize_repeated_volumes = True
    print("Crystal :", crystal.name)
    hc.output_filename = f"test037_test{nb}.root"
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "PreStepUniqueVolumeID",
        "GlobalTime",
    ]

    # singles collection
    sc = sim.add_actor("DigitizerAdderActor", "Singles")
    sc.input_digi_collection = "Hits"
    # sc.policy = "EnergyWinnerPosition"
    sc.policy = "EnergyWeightedCentroidPosition"
    sc.output_filename = hc.output_filename

    return sim


def default_root_hits_branches():
    k1 = ["posX", "posY", "posZ", "edep", "time"]
    k2 = [
        "PostPosition_X",
        "PostPosition_Y",
        "PostPosition_Z",
        "TotalEnergyDeposit",
        "GlobalTime",
    ]
    return k1, k2


def default_root_singles_branches():
    k1 = ["globalPosX", "globalPosY", "globalPosZ", "energy", "time"]
    k2 = [
        "PostPosition_X",
        "PostPosition_Y",
        "PostPosition_Z",
        "TotalEnergyDeposit",
        "GlobalTime",
    ]
    return k1, k2


def check_root_hits(paths, nb, ref_hits_output, hits_output, png_output="auto"):
    if png_output == "auto":
        png_output = f"test037_test{nb}_hits.png"
    # check phsp (new version)
    print()
    gate.exception.warning(f"Check root (hits)")
    k1, k2 = default_root_hits_branches()
    p1 = utility.root_compare_param_tree(ref_hits_output, "Hits", k1)
    # in the legacy gate, some edep=0 are still saved in the root file,
    # so we don't count that ones in the histogram comparison
    p1.mins[k1.index("edep")] = 0
    p2 = utility.root_compare_param_tree(hits_output, "Hits", k2)
    p2.scaling[p2.the_keys.index("GlobalTime")] = 1e-9  # time in ns
    p = utility.root_compare_param(p1.the_keys, paths.output / png_output)
    p.hits_tol = 6  # % tolerance (including the edep zeros)
    p.tols[k1.index("posX")] = 10
    p.tols[k1.index("posY")] = 10
    p.tols[k1.index("posZ")] = 1.5
    p.tols[k1.index("edep")] = 0.002
    p.tols[k1.index("time")] = 0.0001
    is_ok = utility.root_compare4(p1, p2, p)

    return is_ok


def check_root_singles(
    paths, v, ref_singles_output, singles_output, sname="Singles", png_output="auto"
):
    if png_output == "auto":
        png_output = f"test037_test{v}_singles.png"
    # check phsp (singles)
    print()
    gate.exception.warning(f"Check root (singles)")
    k1, k2 = default_root_singles_branches()
    p1 = utility.root_compare_param_tree(ref_singles_output, "Singles", k1)
    # in the legacy gate, some edep=0 are still saved in the root file,
    # so we don't count that ones in the histogram comparison
    p1.mins[k1.index("energy")] = 0
    p2 = utility.root_compare_param_tree(singles_output, sname, k2)
    p2.scaling[p2.the_keys.index("GlobalTime")] = 1e-9  # time in ns
    p = utility.root_compare_param(p1.the_keys, paths.output / png_output)
    p.hits_tol = 5  # % tolerance (including the edep zeros)
    p.tols[k1.index("globalPosX")] = 5
    p.tols[k1.index("globalPosY")] = 5
    p.tols[k1.index("globalPosZ")] = 1.5
    p.tols[k1.index("energy")] = 0.0045
    p.tols[k1.index("time")] = 0.0001

    is_ok = utility.root_compare4(p1, p2, p)

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.tests import utility
import uproot
import numpy as np


def check_scatter(root_filename):
    root_file = uproot.open(root_filename)
    phsp1 = root_file["phsp"]
    phsp_scatter = root_file["phsp_scatter"]
    phsp_no_scatter = root_file["phsp_no_scatter"]

    print(f"Number of entries in phsp = {phsp1.num_entries}")
    print(f"Number of entries in scatter filtered phsp = {phsp_scatter.num_entries}")
    print(
        f"Number of entries in no_scatter filtered phsp = {phsp_no_scatter.num_entries}"
    )

    scatter_flag = phsp1["UnscatteredPrimaryFlag"].array()
    scatter_flag2 = phsp_scatter["UnscatteredPrimaryFlag"].array()
    scatter_flag3 = phsp_no_scatter["UnscatteredPrimaryFlag"].array()

    # Count the entries where UnscatteredPrimaryFlag is 1 or -1
    n_scatter = np.sum(scatter_flag == 1)
    n_no_scatter = np.sum(scatter_flag == 0)

    print(f"phsp1,   nb flag == 0  -> {np.sum(scatter_flag == 0)}")
    print(f"phsp1,   nb flag == 1  -> {np.sum(scatter_flag == 1)}")
    print(f"phsp1,   nb flag == -1 -> {np.sum(scatter_flag == -1)}")
    print()

    print(f"phsp_s,  nb flag == 0  -> {np.sum(scatter_flag2 == 0)}")
    print(f"phsp_s,  nb flag == 1  -> {np.sum(scatter_flag2 == 1)}")
    print(f"phsp_s,  nb flag == -1 -> {np.sum(scatter_flag2 == -1)}")
    print()

    print(f"phsp_ns, nb flag == 0  -> {np.sum(scatter_flag3 == 0)}")
    print(f"phsp_ns, nb flag == 1  -> {np.sum(scatter_flag3 == 1)}")
    print(f"phsp_ns, nb flag == -1 -> {np.sum(scatter_flag3 == -1)}")
    print()

    # check
    # tol = 0.01
    # r = np.abs(n_scatter - phsp_scatter.num_entries) / n_scatter
    b1 = n_scatter == phsp_scatter.num_entries
    utility.print_test(
        b1,
        f"Number of scatter flag is {n_scatter} and "
        f"filtered = {phsp_scatter.num_entries} entries",
    )

    b2 = n_no_scatter == phsp_no_scatter.num_entries
    utility.print_test(
        b2,
        f"Number of no_scatter flag is {n_no_scatter} and "
        f"filtered = {phsp_no_scatter.num_entries} entries",
    )

    b3 = phsp_scatter.num_entries == np.sum(scatter_flag2 == 1)
    utility.print_test(
        b3,
        f"Check nb filter {phsp_scatter.num_entries} vs {np.sum(scatter_flag2 == 1)}",
    )

    b4 = phsp_no_scatter.num_entries == np.sum(scatter_flag3 != 1)
    utility.print_test(
        b4,
        f"Check nb filter {phsp_no_scatter.num_entries} vs {np.sum(scatter_flag3 != 1)}",
    )

    ####
    phsp1_particles = {}
    phsp2_particles = {}
    print(phsp1.keys())
    for particle in phsp1.arrays():
        if particle["UnscatteredPrimaryFlag"] != 1:
            continue
        key = (particle["EventID"], particle["TrackID"])
        phsp1_particles[key] = particle

    for particle in phsp_scatter.arrays():
        key = (particle["EventID"], particle["TrackID"])
        phsp2_particles[key] = particle

    unique_in_phsp1 = set(phsp1_particles.keys()) - set(phsp2_particles.keys())
    unique_in_phsp2 = set(phsp2_particles.keys()) - set(phsp1_particles.keys())

    print("Particles in phsp1 but not in phsp2:")
    i = 0
    for key in unique_in_phsp1:
        print(phsp1_particles[key])
        i += 1
    print(i)

    print("\nParticles in phsp2 but not in phsp1:")
    i = 0
    for key in unique_in_phsp2:
        p = phsp2_particles[key]
        print(
            p,
            p["EventID"],
            p["TrackID"],
            p["ParentID"],
            p["PreDirection_X"],
            p["PreDirection_Y"],
            p["PreDirection_Z"],
            p["UnscatteredPrimaryFlag"],
            p["KineticEnergy"],
        )
        i += 1
    print(i)

    return b1 and b2 and b3 and b4



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate.tests.utility as utility
import opengate_core as g4

if __name__ == "__main__":
    a = "titi"
    s = g4.G4String(a)
    print("G4String =", s, type(s))
    assert s == "titi"
    assert type(s) == g4.G4String

    t = g4.G4String(s)
    print("G4String =", t, type(t))
    assert t == "titi"

    t = "toto"
    print(f"s={s} t={t} ", type(t))

    assert s == "titi"
    assert t == "toto"

    utility.test_ok(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test085_free_flight_helpers import *
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test085_spect"
    )

    # create the simulation
    sim = gate.Simulation()
    sim.number_of_threads = 8
    # im.visu = True
    source, actors = create_simulation_test085(
        sim,
        paths,
        simu_name="ref",
        ac=5e6,
        use_spect_head=True,
        use_spect_arf=False,
        use_phsp=False,
    )

    # go
    sim.run()
    stats = sim.get_actor("stats")
    print(stats)

    # not really a test, generate reference simulation for FF
    utility.test_ok(True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test027_fake_spect", "test027_fake_spect"
    )

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.visu = False
    sim.number_of_threads = 2
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    keV = gate.g4_units.keV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq

    # world size
    sim.world.size = [2 * m, 2 * m, 2 * m]

    # material
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # fake spect head
    waterbox = sim.add_volume("Box", "SPECThead")
    waterbox.size = [55 * cm, 42 * cm, 18 * cm]
    waterbox.material = "G4_AIR"

    # crystal
    crystal = sim.add_volume("Box", "crystal")
    crystal.mother = "SPECThead"
    crystal.size = [55 * cm, 42 * cm, 2 * cm]
    crystal.translation = [0, 0, 4 * cm]
    crystal.material = "NaITl"
    crystal.color = [1, 1, 0, 1]

    # physic list
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 0.01 * mm
    sim.physics_manager.global_production_cuts.electron = 0.01 * mm
    sim.physics_manager.global_production_cuts.positron = 1 * mm
    sim.physics_manager.global_production_cuts.proton = 1 * mm

    # default source for tests
    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.mono = 140.5 * keV
    source.position.type = "sphere"
    source.position.radius = 4 * cm
    source.position.translation = [0, 0, -15 * cm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = 5000 * Bq / sim.number_of_threads

    # add stat actor
    sim.add_actor("SimulationStatisticsActor", "Stats")

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    hc.attached_to = crystal.name
    hc.output_filename = "test027.root"
    hc.attributes = [
        "KineticEnergy",
        "PostPosition",
        "PrePosition",
        "TotalEnergyDeposit",
        "GlobalTime",
        "TrackVolumeName",
        "TrackID",
        "PostStepUniqueVolumeID",
        "PreStepUniqueVolumeID",
        "TrackVolumeCopyNo",
        "TrackVolumeInstanceID",
    ]

    # singles collection
    sc = sim.add_actor("DigitizerAdderActor", "Singles")
    sc.attached_to = crystal.name
    sc.input_digi_collection = "Hits"
    sc.policy = "EnergyWinnerPosition"
    # sc.policy = 'EnergyWeightedCentroidPosition'
    # same filename, there will be two branches in the file
    sc.output_filename = hc.output_filename

    sec = gate.g4_units.second
    sim.running_verbose_level = gate.logger.RUN
    sim.run_timing_intervals = [
        [0, 0.33 * sec],
        [0.33 * sec, 0.66 * sec],
        [0.66 * sec, 1 * sec],
    ]

    # start simulation
    sim.run()

    # stat
    gate.exception.warning("Compare stats")
    stats = sim.get_actor("Stats")
    print(stats)
    print(f"Number of runs was {stats.counts.runs}. Set to 1 before comparison")
    stats.counts.runs = 1  # force to 1
    stats_ref = utility.read_stat_file(paths.gate_output / "stat.txt")
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.07)

    # root compare HITS
    print()
    gate.exception.warning("Compare HITS")
    gate_file = paths.gate_output / "spect.root"
    checked_keys = ["posX", "posY", "posZ", "edep", "time", "trackId"]
    utility.compare_root(
        gate_file,
        hc.get_output_path(),
        "Hits",
        "Hits",
        checked_keys,
        paths.output / "test027.png",
    )

    # Root compare SINGLES
    print()
    gate.exception.warning("Compare SINGLES")
    gate_file = paths.gate_output / "spect.root"
    checked_keys = ["globalposX", "globalposY", "globalposZ", "energy"]
    utility.compare_root(
        gate_file,
        sc.get_output_path(),
        "Singles",
        "Singles",
        checked_keys,
        paths.output / "test027_singles.png",
    )

    # this is the end, my friend
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import uproot
import opengate as gate
from opengate.tests import utility


def create_simulation(sim, paths, name):
    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    nm = gate.g4_units.nm

    sim.output_dir = paths.output

    # waterworld
    world = sim.world
    world.size = [100 * m, 100 * m, 100 * m]
    world.material = "G4_WATER"

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.enable_decay = True
    sim.physics_manager.set_production_cut("world", "all", 1e6 * mm)

    # radionuclide
    z = 89
    a = 225
    s1 = sim.add_source("GenericSource", "source")
    s1.particle = f"ion {z} {a}"
    s1.position.type = "sphere"
    s1.position.radius = 1 * nm
    s1.position.translation = [0, 0, 0]
    s1.direction.type = "iso"
    s1.n = 50

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "stats")
    s.track_types_flag = True

    # phsp actor
    phsp = sim.add_actor("PhaseSpaceActor", "phsp")
    phsp.attributes = [
        "KineticEnergy",
        "EventID",
        "TrackID",
        "ParentID",
        "GlobalTime",
        "ParticleName",
        "ParentParticleName",
        "TrackCreatorProcess",
        "ProcessDefinedStep",
    ]
    phsp.debug = False
    phsp.output_filename = f"test061_{name}.root"
    phsp.steps_to_store = "first"


def analyse(simulation):
    # end
    stats = simulation.actor_manager.get_actor("stats")
    print(stats)

    # open root file
    phsp = simulation.actor_manager.get_actor("phsp")
    root = uproot.open(phsp.get_output_path())
    tree = root[root.keys()[0]]

    print(f"phsp.number_of_absorbed_events: {phsp.number_of_absorbed_events}")

    # Get the arrays from the tree
    events = tree.arrays(
        [
            "GlobalTime",
            "KineticEnergy",
            "ParentParticleName",
            "ParticleName",
            "EventID",
            "TrackID",
            "ParentID",
        ]
    )
    global_time_array = events["GlobalTime"]
    kinetic_energy_array = events["KineticEnergy"]
    parent_particle_array = events["ParentParticleName"]
    particle_array = events["ParticleName"]
    event_id_array = events["EventID"]
    track_id_array = events["TrackID"]
    parent_id_array = events["ParentID"]
    parent_particle_array_ref = ["unknown"] * len(kinetic_energy_array)

    # we test if the stored ParentParticleName is the same as the one we can retrieve from
    # the TrackID and ParentID

    def set_parent_name(parent_by_track, starting_index):
        # print('set parent name', starting_index, len(parent_by_track))
        lindex = starting_index
        lis_ok = True
        while lindex < starting_index + len(parent_by_track):
            pid = parent_id_array[lindex]
            if pid in parent_by_track:
                parent_particle_array_ref[lindex] = parent_by_track[pid]
            else:
                # the events are not stored in the phsp because they do not do a single step,
                # for the test we force to Ac225 ? NO ! was ok in G4 11.1, but no more valid in G4 11.2
                # now we check if this is an ion with [] inside the name
                parent_particle_array_ref[lindex] = "Ac225"
            ok = (
                parent_particle_array_ref[lindex] == parent_particle_array[lindex]
                or "[" in parent_particle_array[lindex]
            )
            if not ok or lindex % 100 == 0:
                utility.print_test(
                    ok,
                    f"Parent particle i={lindex} track={track_id_array[lindex]} name={particle_array[lindex]}  ptrack={pid} "
                    f"       ref = {parent_particle_array_ref[lindex]} "
                    f"vs = {parent_particle_array[lindex]}",
                )
            if not ok:
                lis_ok = False
            lindex += 1
        return lis_ok

    index = 0
    parent_by_track = {}
    current_event_id = -1
    starting_index = 0
    is_ok = True
    while index < tree.num_entries:
        event_id = event_id_array[index]
        if current_event_id != event_id:
            is_ok = set_parent_name(parent_by_track, starting_index) and is_ok
            parent_by_track = {}
            starting_index = index
            current_event_id = event_id
        track_id = track_id_array[index]
        parent_by_track[track_id] = particle_array[index]
        index += 1

    print()
    utility.print_test(is_ok, f"Compared {len(kinetic_energy_array)} elements")
    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.sources.phidsources as phidsources
import click
import matplotlib.pyplot as plt

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("rad_name", nargs=1)
@click.option("--output", "-o", default=None, help="output plot in this file")
@click.option(
    "--verbose/--no-verbose", "-v", is_flag=True, default=False, help="Print data"
)
@click.option("--plot/--no-plot", is_flag=True, default=True, help="Plot data")
@click.option(
    "--iaea_web_site",
    is_flag=True,
    default=False,
    help="Force loading data from IAEA website and store to local file",
)
@click.option(
    "--from_data_file",
    default=None,
    help="Force loading data from the given file and store to local file",
)
@click.option(
    "--log_scale", "--log", is_flag=True, default=False, help="plot with log scale"
)
def go(rad_name, output, plot, verbose, log_scale, iaea_web_site, from_data_file):
    """
    Display the atomic relaxation photon spectra for one given radionuclide.

    Data are retrieved from https://www-nds.iaea.org/relnsd/vcharthtml/VChartHTML.html
    By default, data are stored in text files in the folder: opengate/opengate/data/atomic_relaxation
    """

    # get nuclide
    nuclide = phidsources.get_nuclide_from_name(rad_name)

    # load the atomic relaxation data
    if iaea_web_site and from_data_file is not None:
        gate.fatal(f"Use either --iaea_web_site or --from_data_file, not both")
    load_type = "local"
    if iaea_web_site:
        load_type = "iaea_web_site"
    if from_data_file:
        load_type = from_data_file
    ene, weights = phidsources.atomic_relaxation_load(nuclide, load_type)

    # print
    keV = gate.g4_units.keV
    if verbose:
        i = 0
        for e, w in zip(ene, weights):
            print(f"{i} Energy {e / keV:.3f} keV     {w * 100:.3f} %")
            i += 1
        print()

    # plot
    if plot:
        f, ax = plt.subplots(1, 1, figsize=(15, 5))
        ax.bar(
            ene / keV,
            height=weights * 100,
            width=0.2,
            label=f"{nuclide.nuclide}",
            log=log_scale,
        )
        ax.set_xlabel("Energy in keV")
        ax.set_ylabel("Intensity in %")
        ax.legend()
        if output:
            print(f"Output plot in {output}")
            f.savefig(output)
        else:
            plt.show()


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.contrib.spect.spect_config import *
from opengate.tests import utility

if __name__ == "__main__":

    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test094_pytomography"
    )

    data_folder = paths.data
    output_folder = paths.output  # will be copied in the reference folder
    number_of_threads = 4

    # units
    deg = g4_units.deg
    cm = g4_units.cm
    mm = g4_units.mm
    Bq = g4_units.Bq

    # Init SPECT simulation
    sc = SPECTConfig()
    sc.output_folder = output_folder
    sc.number_of_threads = number_of_threads
    sc.detector_config.model = "intevo"
    sc.detector_config.collimator = "melp"
    sc.detector_config.number_of_heads = 2
    sc.detector_config.size = [128, 128]
    sc.detector_config.spacing = [4.8 * mm, 4.8 * mm]
    sc.detector_config.digitizer_function = intevo.add_digitizer
    sc.detector_config.digitizer_channels = get_default_energy_windows("lu177")
    sc.phantom_config.image = data_folder / "ct_5mm.mhd"
    sc.phantom_config.translation = [0, 0, -20 * g4_units.cm]
    sc.source_config.image = data_folder / "3_sources_5mm_v2.mhd"
    sc.source_config.radionuclide = "lu177"
    sc.acquisition_config.radius = 420 * g4_units.mm
    sc.acquisition_config.duration = 30 * g4_units.s
    sc.acquisition_config.number_of_angles = 30

    sc.free_flight_config.primary_activity = 1e6 * Bq
    sc.free_flight_config.scatter_activity = 2e6 * Bq
    sc.free_flight_config.angle_tolerance = 15 * deg
    sc.free_flight_config.forced_direction_flag = True
    sc.free_flight_config.angle_tolerance_min_distance = 6 * cm
    sc.free_flight_config.max_compton_level = 5
    sc.free_flight_config.compton_splitting_factor = 50
    sc.free_flight_config.rayleigh_splitting_factor = 50

    # debug
    # sc.free_flight_config.primary_activity = 5e1 * Bq
    # sc.free_flight_config.scatter_activity = 5e1 * Bq
    # sc.acquisition_config.number_of_angles = 1

    # run 1: primary
    sim = gate.Simulation()
    sc.setup_simulation_ff_primary(sim, visu=False)
    sim.run(start_new_process=True)
    stats = sim.find_actors("stats")[0]
    print(stats)

    # run 2: scatter
    print()
    sim = gate.Simulation()
    sc.setup_simulation_ff_scatter(sim, visu=False)
    sim.run(start_new_process=True)
    stats = sim.find_actors("stats")[0]
    print(stats)

    # combine and compute relative uncertainty
    n_prim = sc.free_flight_config.primary_activity / Bq
    n_scatter = sc.free_flight_config.scatter_activity / Bq
    n_ref = 1e8
    spect_freeflight_merge_all_heads(
        output_folder, n_prim, n_scatter, n_ref, nb_of_heads=2, verbose=True
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test039_hits_memory_check_helpers as test39
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, output_folder="test039")

    # create the simulation
    sim = test39.create_simu(8, paths)
    sim.random_seed = 987654321

    # go
    sim.run()

    is_ok = test39.test_results(sim, paths)

    utility.test_ok(is_ok)



from box import Box
import numpy as np
import itk
import threading

import opengate_core as g4
from ..utility import g4_units, LazyModuleLoader
from ..exception import fatal
from .base import ActorBase
from .digitizers import (
    DigitizerEnergyWindowsActor,
)
from .actoroutput import ActorOutputSingleImage, ActorOutputRoot
from ..base import process_cls

garf = LazyModuleLoader("garf")


def check_channel_overlap(ch1, ch2):
    ch1 = Box(ch1)
    ch2 = Box(ch2)
    if ch2.min < ch1.min < ch2.max:
        return True
    if ch2.min < ch1.max < ch2.max:
        return True
    if ch1.min < ch2.min < ch1.max:
        return True
    if ch1.min < ch2.max < ch1.max:
        return True
    return False


class ARFTrainingDatasetActor(ActorBase, g4.GateARFTrainingDatasetActor):
    """
    The ARFTrainingDatasetActor build a root file with energy, angles, positions and energy windows
    of a spect detector. To be used by garf_train to train a ARF neural network.

    Note: Must inherit from ActorBase not from HitsCollectionActor, even if the
    cpp part inherit from HitsCollectionActor
    """

    user_info_defaults = {
        "attributes": (
            [],
            {
                "doc": "FIXME",
            },
        ),
        "debug": (False, {"doc": "FIXME"}),
        "energy_windows_actor": (
            None,
            {
                "doc": "",
            },
        ),
        # duplicated because cpp part inherit from HitsCollectionActor
        "clear_every": (
            1e5,
            {
                "doc": "FIXME",
            },
        ),
        "plane_axis": (
            [0, 1, 2],
            {
                "doc": "Axe indices for the plane",
            },
        ),
        # duplicated because cpp part inherit from HitsCollectionActor
        "keep_zero_edep": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "russian_roulette": (1, {"doc": "Russian roulette factor. "}),
    }

    user_output_config = {
        "root_output": {
            "actor_output_class": ActorOutputRoot,
        },
    }

    def __init__(self, *args, **kwargs):
        ActorBase.__init__(self, *args, **kwargs)
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateARFTrainingDatasetActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "SteppingAction",
                # "BeginOfRunActionMasterThread",
                # "EndOfRunActionMasterThread",
                "BeginOfEventAction",
                "EndOfEventAction",
            }
        )

    def initialize(self):
        ActorBase.initialize(self)
        self.check_energy_window_actor()
        # initialize C++ side
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()

    def StartSimulationAction(self):
        g4.GateARFTrainingDatasetActor.StartSimulationAction(self)

    def check_energy_window_actor(self):
        # check the energy_windows_actor
        if not self.energy_windows_actor in self.simulation.actor_manager.actors:
            fatal(
                f"The actor '{self.name}' has the user input energy_windows_actor={self.energy_windows_actor}, "
                f"but no actor with this name was found in the simulation."
            )
        ewa = self.simulation.get_actor(self.energy_windows_actor)
        if not isinstance(ewa, DigitizerEnergyWindowsActor):
            fatal(
                f"The actor '{self.name}' has the user input energy_windows_actor={self.energy_windows_actor}, "
                f"but {ewa.name} is not the correct type of actor. "
                f"It should be a DigitizerEnergyWindowsActor, while it is a {type(ewa).__name}. "
            )
        # check overlap in channels
        channels = ewa.channels
        for i, ch1 in enumerate(channels):
            for j, ch2 in enumerate(channels[i + 1 :]):
                is_overlap = check_channel_overlap(ch1, ch2)
                if is_overlap:
                    fatal(
                        f'In the actor "{self.user_info.name}", the energy channels '
                        f"{i} and {i + 1 + j} overlap. This is not possible for ARF. \n"
                        f"{ch1} {ch2}"
                    )

    def EndSimulationAction(self):
        g4.GateARFTrainingDatasetActor.EndSimulationAction(self)
        ActorBase.EndSimulationAction(self)


def _setter_hook_image_spacing(self, image_spacing):
    # force float
    return [float(s) for s in image_spacing]


class ARFActor(ActorBase, g4.GateARFActor):
    """
    The ARF Actor is attached to a volume.
    Every time a particle reaches the volume, it considers the energy and the direction of the particle.
    It runs the neural network model to provide the probability of detection in all energy windows.

    Output is an ITK image that can be retrieved with self.output_image
    """

    user_info_defaults = {
        "batch_size": (
            2e5,
            {
                "doc": "FIXME",
            },
        ),
        "pth_filename": (
            None,
            {
                "doc": "FIXME",
            },
        ),
        "image_size": (
            [128, 128],
            {
                "doc": "FIXME",
            },
        ),
        "image_spacing": (
            [4.41806 * g4_units.mm, 4.41806 * g4_units.mm],
            {"doc": "FIXME", "setter_hook": _setter_hook_image_spacing},
        ),
        "distance_to_crystal": (
            75 * g4_units.mm,
            {
                "doc": "FIXME",
            },
        ),
        "verbose_batch": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "enable_hit_slice": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "flip_plane": (
            False,
            {
                "doc": "FIXME",
            },
        ),
        "plane_axis": (
            [0, 1, 2],
            {
                "doc": "Axe indices for the plane",
            },
        ),
        "gpu_mode": (
            "auto",
            {"doc": "FIXME", "allowed_values": ("cpu", "gpu", "auto")},
        ),
    }

    user_output_config = {
        "counts": {
            "actor_output_class": ActorOutputSingleImage,
        },
        "squared_counts": {
            "actor_output_class": ActorOutputSingleImage,
            "active": False,
        },
    }

    def __init__(self, *args, **kwargs):
        ActorBase.__init__(self, *args, **kwargs)
        # import module
        self.debug_nb_hits_before = None
        self.debug_nb_hits = 0
        # prepare output
        self.nn = None
        self.model = None
        self.model_data = None
        self.batch_nb = 0
        self.detected_particles = 0
        # need a lock when the ARF is applied
        self.lock = None
        # local variables
        self.image_plane_spacing = None
        self.image_plane_size_pixel = None
        self.image_plane_size_mm = None
        self.output_image = None
        self.output_array = None
        self.output_squared_array = None
        self.output_size = None
        self.nb_ene = None

        # self._add_user_output(ActorOutputSingleImage, "arf_projection")
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateARFActor.__init__(self, self.user_info)
        self.AddActions(
            {
                "SteppingAction",
                "BeginOfRunActionMasterThread",
                "EndOfRunActionMasterThread",
                "BeginOfRunAction",
                "EndOfRunAction",
            }
        )

    def __getstate__(self):
        # needed to not pickle objects that cannot be pickled (g4, cuda, lock, etc).
        return_dict = super().__getstate__()
        return_dict["nn"] = None
        return_dict["lock"] = None
        return_dict["model"] = None
        return return_dict

    def initialize(self):
        # call the initialize() method from the super class (python-side)
        ActorBase.initialize(self)
        self.lock = threading.Lock()

        self.debug_nb_hits_before = 0
        self.debug_nb_hits = 0

        self.initialize_model()
        self.initialize_params()
        self.initialize_device()

        self.output_array = np.zeros(self.output_size, dtype=np.float64)

        # squared image ?
        if self.user_output.squared_counts.get_active():
            self.output_squared_array = np.zeros(self.output_size, dtype=np.float64)

        # initialize C++ side
        self.InitializeUserInfo(self.user_info)
        self.InitializeCpp()
        self.SetARFFunction(self.apply)

    def initialize_model(self):
        # load the pth file
        self.nn, self.model = garf.load_nn(
            self.pth_filename, verbose=False, gpu_mode=self.gpu_mode
        )

        # size and spacing (2D)
        self.image_plane_spacing = np.array(
            [self.user_info.image_spacing[0], self.user_info.image_spacing[1]]
        )
        self.image_plane_size_pixel = np.array(
            [self.user_info.image_size[0], self.user_info.image_size[1]]
        )
        self.image_plane_size_mm = (
            self.image_plane_size_pixel * self.image_plane_spacing
        )

        # shortcut to model_data
        self.model_data = self.nn["model_data"]

    def initialize_device(self):
        # which device for GARF: cpu cuda mps?
        current_gpu_mode, current_gpu_device = garf.helpers.get_gpu_device(
            self.gpu_mode
        )
        self.model_data["current_gpu_device"] = current_gpu_device
        self.model_data["current_gpu_mode"] = current_gpu_mode
        self.model.to(current_gpu_device)

    def initialize_params(self):
        # output image: nb of energy windows times nb of runs (for rotation)
        self.nb_ene = self.model_data["n_ene_win"]
        nb_runs = len(self.simulation.run_timing_intervals)

        if not self.enable_hit_slice:
            self.nb_ene -= 1

        # size and spacing in 3D
        self.output_image = np.array(
            [
                self.nb_ene,
                self.image_plane_size_pixel[0],
                self.image_plane_size_pixel[1],
            ]
        )
        self.output_size = [
            self.nb_ene * nb_runs,
            self.output_image[1],
            self.output_image[2],
        ]
        self.output_image = np.zeros(self.output_size, dtype=np.float64)

    def apply(self, actor):
        # we need a lock when the ARF is applied
        if self.simulation.use_multithread:
            with self.lock:
                self.arf_build_image_from_projected_points(actor)
        else:
            self.arf_build_image_from_projected_points(actor)

    def arf_build_image_from_projected_points(self, actor):
        # get values from the cpp side
        energy = np.array(actor.GetEnergy())
        pos_x = np.array(actor.GetPositionX())
        pos_y = np.array(actor.GetPositionY())
        dir_x = np.array(actor.GetDirectionX())
        dir_y = np.array(actor.GetDirectionY())
        dir_z = np.array(actor.GetDirectionZ())
        weights = np.array(actor.GetWeights())

        # do nothing if no hits
        if energy.size == 0:
            return

        # (do NOT use plane_axis here, it is included in GateARFActor)

        # build the data by passing direction vectors, NOT angles
        px_base = (pos_x, pos_y, dir_x, dir_y, dir_z, energy)
        if len(weights) == 0:
            px = np.column_stack(px_base)
        else:
            px = np.column_stack(px_base + (weights,))
        self.debug_nb_hits_before += len(px)

        # verbose current batch
        if self.verbose_batch:
            print(
                f"Apply ARF to {energy.shape[0]} hits (device = {self.model_data['current_gpu_mode']})"
            )

        # from projected points to image counts
        u, v, w_pred = garf.arf_from_points_to_image_counts(
            px,
            self.model,
            self.model_data,
            self.user_info.distance_to_crystal,
            self.image_plane_size_mm,
            self.image_plane_size_pixel,
            self.image_plane_spacing,
        )

        # do nothing if there is no hit in the image
        if u.shape[0] != 0:
            run_id = actor.GetCurrentRunId()
            s = self.nb_ene * run_id
            img = self.output_array[s : s + self.nb_ene]
            garf.image_from_coordinates_add_numpy(
                img, u, v, w_pred, self.enable_hit_slice
            )
            # squared ?
            if self.user_output.squared_counts.get_active():
                w_pred_squared = w_pred * w_pred
                img = self.output_squared_array[s : s + self.nb_ene]
                garf.image_from_coordinates_add_numpy(
                    img, u, v, w_pred_squared, self.enable_hit_slice
                )
            self.debug_nb_hits += u.shape[0]

    def EndOfRunActionMasterThread(self, run_index):
        nb_slice = self.nb_ene

        # convert to itk image
        # FIXME: this should probably go into EndOfRunAction
        output_image = itk.image_from_array(self.output_array)

        # set spacing and origin like DigitizerProjectionActor
        spacing = self.image_spacing
        spacing = np.array([spacing[0], spacing[1], 1])
        size = np.array([0, 0, 0])
        size[0] = self.image_plane_size_pixel[0]
        size[1] = self.image_plane_size_pixel[1]
        size[2] = nb_slice
        origin = -size / 2.0 * spacing + spacing / 2.0
        origin[2] = 0
        output_image.SetSpacing(spacing)
        output_image.SetOrigin(origin)

        # convert double to float
        InputImageType = itk.Image[itk.D, 3]
        OutputImageType = itk.Image[itk.F, 3]
        castImageFilter = itk.CastImageFilter[InputImageType, OutputImageType].New()
        castImageFilter.SetInput(output_image)
        castImageFilter.Update()
        output_image = castImageFilter.GetOutput()
        self.user_output["counts"].store_data("merged", output_image)

        # squared ?
        if self.user_output.squared_counts.get_active():
            output_image = itk.image_from_array(self.output_squared_array)
            castImageFilter = itk.CastImageFilter[InputImageType, OutputImageType].New()
            castImageFilter.SetInput(output_image)
            castImageFilter.Update()
            output_image = castImageFilter.GetOutput()
            self.user_output["squared_counts"].store_data("merged", output_image)

        # unsure why return 0?
        return 0

    def EndSimulationAction(self):
        g4.GateARFActor.EndSimulationAction(self)
        ActorBase.EndSimulationAction(self)
        # process the remaining elements in the batch
        # self.apply()
        # warning('SHOULD call apply here ???')
        self.user_output["counts"].write_data_if_requested(which="merged")
        if self.user_output.squared_counts.get_active():
            self.user_output["squared_counts"].write_data_if_requested(which="merged")


process_cls(ARFActor)
process_cls(ARFTrainingDatasetActor)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.phantoms.nemaiec as gate_iec
from opengate.tests import utility


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", "test038")

    # create the simulation
    sim = gate.Simulation()

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    cm3 = gate.g4_units.cm3
    keV = gate.g4_units.keV
    mm = gate.g4_units.mm
    Bq = gate.g4_units.Bq
    BqmL = Bq / cm3
    sec = gate.g4_units.second
    deg = gate.g4_units.deg
    kBq = 1000 * Bq
    MBq = 1000 * kBq

    # main parameters
    sim.check_volumes_overlap = True
    sim.number_of_threads = 2
    sim.random_seed = 321654987
    sim.output_dir = paths.output
    ac = 100 * BqmL / sim.number_of_threads
    sim.visu = False
    if sim.visu:
        ac = 10 * BqmL  # per mL
        sim.number_of_threads = 1

    # world size
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]
    world.material = "G4_AIR"

    # iec phantom
    iec_phantom = gate_iec.add_iec_phantom(sim)

    # cylinder for phsp
    sph_surface = sim.add_volume("Sphere", "phase_space_sphere")
    sph_surface.rmin = 210 * mm
    sph_surface.rmax = 211 * mm
    sph_surface.color = [0, 1, 0, 1]
    sph_surface.material = "G4_AIR"

    # physic list
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option4"
    sim.physics_manager.set_production_cut("world", "all", 1 * mm)

    # source sphere
    gate_iec.add_spheres_sources(
        sim,
        "iec",
        "source1",
        [10, 13, 17, 22, 28, 37],
        [ac, ac, ac, ac, ac, ac],
        verbose=True,
    )

    sources = sim.source_manager.sources
    for source in sources.values():
        source.particle = "gamma"
        source.energy.type = "mono"
        source.energy.mono = 140.5 * keV

    # background source 1:10 ratio with sphere
    bg = gate_iec.add_background_source(sim, "iec", "source_bg", ac / 7, verbose=True)
    bg.particle = "gamma"
    bg.energy.type = "mono"
    bg.energy.mono = 140.5 * keV

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats.output_filename = "test038_train_stats.txt"

    # filter gamma only
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "gamma"

    # phsp
    phsp = sim.add_actor("PhaseSpaceActor", "phase_space")
    phsp.attached_to = "phase_space_sphere"
    # we use PrePosition because this is the first step in the volume
    phsp.attributes = [
        "KineticEnergy",
        "PrePosition",
        "PreDirection",
        "TimeFromBeginOfEvent",  # not needed ; only to test with ideal reconstruction
        # needed for gan_flag
        "EventID",
        "EventKineticEnergy",
        # for conditional :
        "EventPosition",
        "EventDirection",
    ]
    phsp.output_filename = "test038_train.root"
    # this option allow to store all events even if absorbed
    phsp.store_absorbed_event = True
    phsp.filters.append(f)
    print(phsp)
    print(phsp.get_output_path())

    # go
    sim.run(start_new_process=False)

    # ----------------------------------------------------------------------------------------------------------

    # check stats
    print()
    gate.exception.warning(f"Check stats")
    stats = sim.get_actor("Stats")
    print(stats)
    stats_ref = utility.read_stat_file(paths.output_ref / "test038_train_stats.txt")
    stats.counts.runs = 1
    is_ok = utility.assert_stats(stats, stats_ref, 0.02)

    # check phsp
    print()
    gate.exception.warning(f"Check root")
    p = sim.get_actor("phase_space")
    print(f"Number of absorbed : {p.number_of_absorbed_events}")
    ref_file = paths.output_ref / "test038_train.root"
    hc_file = phsp.get_output_path()
    checked_keys = [
        "TimeFromBeginOfEvent",
        "KineticEnergy",
        "PrePosition_X",
        "PrePosition_Y",
        "PrePosition_Z",
        "PreDirection_X",
        "PreDirection_Y",
        "PreDirection_Z",
        "EventKineticEnergy",
        "EventPosition_X",
        "EventPosition_Y",
        "EventPosition_Z",
        "EventDirection_X",
        "EventDirection_Y",
        "EventDirection_Z",
    ]
    scalings = [1] * len(checked_keys)
    # scalings[0] = 1e-9  # time in ns
    tols = [1.0] * len(checked_keys)
    tols[checked_keys.index("TimeFromBeginOfEvent")] = 0.007
    tols[checked_keys.index("KineticEnergy")] = 0.001
    tols[checked_keys.index("PrePosition_X")] = 1.6
    tols[checked_keys.index("PrePosition_Y")] = 1.6
    tols[checked_keys.index("PrePosition_Z")] = 1.6
    tols[checked_keys.index("PreDirection_X")] = 0.007
    tols[checked_keys.index("PreDirection_Y")] = 0.007
    tols[checked_keys.index("PreDirection_Z")] = 0.007
    tols[checked_keys.index("EventKineticEnergy")] = 0.0001
    tols[checked_keys.index("EventPosition_X")] = 1.0
    tols[checked_keys.index("EventPosition_Y")] = 1.0
    tols[checked_keys.index("EventPosition_Z")] = 1.0
    tols[checked_keys.index("EventDirection_X")] = 0.01
    tols[checked_keys.index("EventDirection_Y")] = 0.01
    tols[checked_keys.index("EventDirection_Z")] = 0.01
    print(scalings, tols)
    is_ok = (
        utility.compare_root3(
            ref_file,
            hc_file,
            "phase_space",
            "phase_space",
            checked_keys,
            checked_keys,
            tols,
            scalings,
            scalings,
            paths.output / "test038_train_phsp.png",
        )
        and is_ok
    )

    utility.test_ok(is_ok)



import itk
import numpy as np
from box import Box
from scipy.spatial.transform import Rotation
import math

from .exception import fatal
from .geometry.utility import (
    get_transform_world_to_local,
    vec_g4_as_np,
)
from .definitions import __gate_list_objects__


def update_image_py_to_cpp(py_img, cpp_img, copy_data=False):
    cpp_img.set_size(py_img.GetLargestPossibleRegion().GetSize())
    cpp_img.set_spacing(py_img.GetSpacing())
    cpp_img.set_origin(py_img.GetOrigin())
    # this is needed !
    cpp_img.set_region(
        py_img.GetLargestPossibleRegion().GetIndex(),
        py_img.GetLargestPossibleRegion().GetSize(),
    )
    # It is really a pain to convert GetDirection into
    # something that can be read by SetDirection !
    d = py_img.GetDirection().GetVnlMatrix().as_matrix()
    rotation = itk.GetArrayFromVnlMatrix(d)
    cpp_img.set_direction(rotation)
    if copy_data:
        # FIXME: do we need to return arr to keep reference ?
        # (on cpp side, a copy is made while it should not be needed)
        arr = itk.array_view_from_image(py_img)
        cpp_img.from_pyarray(arr)


def itk_dir_to_rotation(dir):
    return itk.GetArrayFromVnlMatrix(dir.GetVnlMatrix().as_matrix())


def create_3d_image(
    size, spacing, origin=None, pixel_type="float", allocate=True, fill_value=0
):
    dim = 3
    pixel_type = itk.ctype(pixel_type)
    image_type = itk.Image[pixel_type, dim]
    img = image_type.New()
    region = itk.ImageRegion[dim]()
    region.SetSize([int(s) for s in size])
    region.SetIndex([0, 0, 0])
    # spacing = np.array(spacing)
    img.SetRegions(region)
    img.SetSpacing(spacing)
    if origin is not None:
        img.SetOrigin(origin)
    # (default direction)
    if allocate:
        img.Allocate()
        img.FillBuffer(fill_value)
    return img


def create_image_like(like_image, allocate=True, pixel_type=""):
    # TODO fix pixel_type -> copy from image rather than argument
    info = get_info_from_image(like_image)

    if pixel_type:
        img = create_3d_image(
            info.size, info.spacing, pixel_type=pixel_type, allocate=allocate
        )
    else:
        img = create_3d_image(info.size, info.spacing, allocate=allocate)
    img.SetOrigin(info.origin)
    img.SetDirection(info.dir)
    return img


def create_image_like_info(info, allocate=True):
    img = create_3d_image(info.size, info.spacing, allocate=allocate)
    img.SetOrigin(info.origin)
    img.SetDirection(info.dir)
    return img


def get_info_from_image(image):
    info = Box()
    info.size = np.array(itk.size(image)).astype(int)
    info.spacing = np.array(image.GetSpacing())
    info.origin = np.array(image.GetOrigin())
    info.dir = image.GetDirection()
    return info


def read_image_info(path_to_image):
    path_to_image = str(path_to_image)
    image_IO = itk.ImageIOFactory.CreateImageIO(
        path_to_image, itk.CommonEnums.IOFileMode_ReadMode
    )
    if not image_IO:
        fatal(f"Cannot read the image file (itk): {path_to_image}")
    image_IO.SetFileName(path_to_image)
    image_IO.ReadImageInformation()
    info = Box()
    info.filename = path_to_image
    n = info.size = image_IO.GetNumberOfDimensions()
    info.size = np.ones(n).astype(int)
    info.spacing = np.ones(n)
    info.origin = np.ones(n)
    info.dir = np.ones((n, n))
    for i in range(n):
        info.size[i] = image_IO.GetDimensions(i)
        info.spacing[i] = image_IO.GetSpacing(i)
        info.origin[i] = image_IO.GetOrigin(i)
        info.dir[i] = image_IO.GetDirection(i)
    return info


def get_translation_between_images_center(img_name1, img_name2):
    """
    The two images are considered in the same physical space (coordinate system).
    This function computes the translation between their centers.
    Warning, the ITK image origin consider the center of the first voxel, we thus
    consider half a pixel shift for the center.
    """
    info1 = read_image_info(img_name1)
    info2 = read_image_info(img_name2)
    # get the center of the first image in img coordinate system
    center1 = info1.size / 2.0 * info1.spacing + info1.origin - info1.spacing / 2.0
    # get the center of the second image in img coordinate system
    center2 = info2.size / 2.0 * info2.spacing + info2.origin - info2.spacing / 2.0
    return center2 - center1


def get_translation_to_isocenter(img_filename):
    info = read_image_info(img_filename)
    tr = info.size * info.spacing / 2.0 + info.origin
    return tr


def get_origin_wrt_images_g4_position(img_info1, img_info2, translation):
    """
    The two images are considered in the same GATE physical space (coordinate system), so according to the
    centers of both images (+translation).
    This function computes the origin for the second image such as the two images will be in the same
    physical space of the first image.
    Warning, the ITK image origin considers the center of the first voxel, we thus
    consider half a pixel shift for the center.
    """
    half_size1 = img_info1.size * img_info1.spacing / 2.0
    half_size2 = img_info2.size * img_info2.spacing / 2.0
    origin = (
        img_info1.origin
        + half_size1
        - half_size2
        + translation
        - img_info1.spacing / 2.0
        + img_info2.spacing / 2
    )
    return origin


def get_py_image_from_cpp_image(cpp_image, view=True):
    arr = cpp_image.to_pyarray()
    image = itk_image_from_array(arr, view=view)
    image.SetOrigin(cpp_image.origin())
    image.SetSpacing(cpp_image.spacing())
    return image


def itk_image_from_array(arr, view=True):
    """
    When the input numpy array is of shape [1,1,x], the conversion to itk image fails:
    the output image size is with the wrong dimensions.
    We thus 'patch' itk.image_view_from_array to correct the size.

    Not fully sure if this is the way to go.
    """
    if view is True:
        image = itk.image_view_from_array(arr)
    else:
        image = itk.image_from_array(arr)
    if len(arr.shape) == 3 and arr.shape[1] == arr.shape[2] == 1:
        new_region = itk.ImageRegion[3]()
        new_region.SetSize([1, 1, arr.shape[0]])
        image.SetRegions(new_region)
        image.Update()
    return image


def get_image_center(image):
    info = read_image_info(image)
    center = info.size * info.spacing / 2.0  # + info.spacing / 2.0
    return center


def align_image_with_physical_volume(
    volume,
    image,
    initial_translation=None,
    initial_rotation=Rotation.identity().as_matrix(),
    copy_index=0,
):
    if initial_translation is None:
        initial_translation = [0, 0, 0]
    # FIXME rotation not implemented yet
    # get transform from world
    translation, rotation = get_transform_world_to_local(volume, copy_index)
    # compute origin
    info = get_info_from_image(image)
    origin = -info.size * info.spacing / 2.0 + info.spacing / 2.0 + initial_translation
    origin = Rotation.from_matrix(rotation).apply(origin) + translation
    # set origin and direction
    image.SetOrigin(origin)
    image.SetDirection(rotation)


def create_image_with_extent(extent, spacing=(1, 1, 1), margin=0):
    # define the new size and spacing
    spacing = np.array(spacing).astype(float)
    size = np.ceil((extent[1] - extent[0]) / spacing).astype(int) + 2 * margin

    # create image
    image = create_3d_image(size, spacing)

    # The origin is considered to be at the center of first pixel.
    # It is set such that the image is at the exact extent (bounding volume).
    # The volume contour thus goes through the center of the first pixel.
    origin = extent[0] + spacing / 2.0 - margin
    image.SetOrigin(origin)
    return image


def create_image_with_volume_extent(volume, spacing=(1, 1, 1), margin=0):
    if not isinstance(volume, __gate_list_objects__):
        volume = [volume]

    p_min = []
    p_max = []
    for vol in volume:
        pMin_g4vec, pMax_g4vec = vol.bounding_limits
        p_min.append(
            vec_g4_as_np(pMin_g4vec) + vol.translation_list[0]
        )  # FIXME: make this work in case if repeated volumes
        p_max.append(vec_g4_as_np(pMax_g4vec) + vol.translation_list[0])

    extent_lower = np.min(p_min, axis=0)
    extent_upper = np.max(p_max, axis=0)

    return create_image_with_extent((extent_lower, extent_upper), spacing, margin)


def transform_images_point(p, img1, img2):
    index = img1.TransformPhysicalPointToIndex(p)
    pbis = img2.TransformIndexToPhysicalPoint(index)
    return [i for i in pbis]


def compute_image_3D_CDF(image):
    """
    Compute the three CDF (Cumulative Density Function) for the given image
    Warning; numpy order is ZYX

    :param image: itk image
    """
    # consider the image as a np array
    # WARNING: cannot use array_view_from_image here because it seems to lead to
    # a segfault in some cases (linux + multithread + multiple sources)
    # It is unclear why, but for now we use array_from_image that copy the data
    # array = itk.array_view_from_image(image)
    array = itk.array_from_image(image)

    # normalize
    array = array / np.sum(array)

    # Sum image on a single plane along X axis
    sumx = np.sum(array, axis=2)
    # Y axis, sum plane on a single axis along Y axis
    sumxy = np.sum(sumx, axis=1)
    # X 3D CDF
    cdf_x = []
    for i in range(array.shape[0]):  # Z
        cdf_x.append([])
        for j in range(array.shape[1]):  # Y
            # cumulated sum along X axis
            t = np.cumsum(array[i][j])
            # normalise if the last value (sum) is not zero
            if t[-1] != 0:
                t = t / t[-1]
            cdf_x[i].append(t)

    # Y 2D CDF
    cdf_y = []
    for i in range(len(sumx)):  # Z
        t = np.cumsum(sumx[i])
        if t[-1] != 0:
            t = t / t[-1]
        cdf_y.append(t)

    # Z 1D CDF
    cdf_z = np.cumsum(sumxy) / np.sum(sumxy)

    # return
    return cdf_x, cdf_y, cdf_z


def scale_itk_image(img, scale):
    imgarr = itk.array_from_image(img)
    imgarr = imgarr * scale
    # this is important to use the corrected function to deal with 1D images
    # img2 = itk.image_from_array(imgarr)
    img2 = itk_image_from_array(imgarr)
    img2.CopyInformation(img)
    return img2


def divide_itk_images(
    img1_numerator, img2_denominator, filterVal=0, replaceFilteredVal=0
):
    imgarr1 = itk.array_view_from_image(img1_numerator)
    imgarr2 = itk.array_view_from_image(img2_denominator)
    # print(imgarr1.dtype, imgarr1.dtype)
    if imgarr1.shape != imgarr2.shape:
        fatal(
            f"Cannot divide images of different shape. Found {imgarr1.shape} vs. {imgarr2.shape}."
        )
    imgarrOut = imgarr1.copy()
    L_filterInv = imgarr2 != filterVal
    imgarrOut[L_filterInv] = np.divide(imgarr1[L_filterInv], imgarr2[L_filterInv])

    imgarrOut[np.invert(L_filterInv)] = replaceFilteredVal
    imgarrOut = itk_image_from_array(imgarrOut)
    imgarrOut.CopyInformation(img1_numerator)
    return imgarrOut


def sum_itk_images(itk_image_list):
    if not itk_image_list:
        raise ValueError("The image list is empty.")
    summed_image = itk.GetArrayFromImage(itk_image_list[0])
    for itk_image in itk_image_list[1:]:
        array = itk.GetArrayFromImage(itk_image)
        summed_image = np.add(summed_image, array)
    image = itk_image_from_array(summed_image)
    image.CopyInformation(itk_image_list[0])
    return image


def add_constant_to_itk_image(img, constant):
    imgarr = itk.array_view_from_image(img).copy()
    imgarr += constant
    img2 = itk_image_from_array(imgarr)
    img2.CopyInformation(img)
    return img2


def multiply_itk_images(images):
    image_type = type(images[0])
    multiply_image_filter = itk.MultiplyImageFilter[
        image_type, image_type, image_type
    ].New()
    output = images[0]
    for img in images[1:]:
        multiply_image_filter.SetInput1(output)
        multiply_image_filter.SetInput2(img)
        multiply_image_filter.Update()
        output = multiply_image_filter.GetOutput()
    return output


def split_spect_projections(input_filenames, nb_ene):
    """
    The inputs are filenames of several images containing projections for a given spect head
    Each image is composed of nb_ene energy windows and XX angles.
    The number of angles is found by looking at the number of slices.

    The function computes nb_ene itk image with all angles and all heads merged into a list of
    projections stored as a 3D image, to make it easy to reconstruct with RTK.

    """
    nb_heads = len(input_filenames)

    # read the first image to get information
    img = itk.imread(str(input_filenames[0]))
    info = get_info_from_image(img)
    imga = itk.array_view_from_image(img)

    nb_runs = imga.shape[0] // nb_ene
    nb_angles = nb_heads * nb_runs
    # print('Number of heads', nb_heads)
    # print('Number of E windows', nb_ene)
    # print('Number of run', nb_runs)
    # print('Number of angles', nb_angles)

    # create and allocate final images
    outputs_img = []
    outputs_arr = []
    size = [imga.shape[1], imga.shape[2], nb_angles]
    spacing = info.spacing
    for e in range(nb_ene):
        image = create_3d_image(size, spacing)
        image.SetOrigin(img.GetOrigin())
        outputs_img.append(image)
        outputs_arr.append(itk.array_view_from_image(image))

    # loop on heads and create images
    s2 = 0
    for head in range(nb_heads):
        img = itk.imread(str(input_filenames[head]))
        imga = itk.array_view_from_image(img)
        e = 0
        for s in range(imga.shape[0]):
            outputs_arr[e][s2] = imga[s]
            e += 1
            if e >= nb_ene:
                e = 0
                s2 += 1

    # end
    return outputs_img


def compare_itk_image_info(image1, image2):
    are_origins_equal = np.allclose(image1.GetOrigin(), image2.GetOrigin())
    are_spacings_equal = np.allclose(image1.GetSpacing(), image2.GetSpacing())
    are_directions_equal = np.allclose(image1.GetDirection(), image2.GetDirection())
    return are_spacings_equal and are_directions_equal and are_origins_equal


def compare_itk_image_content(image1, image2):
    arr1 = itk.array_from_image(image1)
    arr2 = itk.array_from_image(image2)
    return np.array_equal(arr1, arr2)


def compare_itk_image(filename1, filename2):
    im1 = itk.imread(filename1)
    im2 = itk.imread(filename2)
    return compare_itk_image_info(im1, im2) and compare_itk_image_content(im1, im2)


def write_itk_image(img, file_path):
    # TODO: check if filepath exists
    # TODO: add metadata to file header
    itk.imwrite(img, str(file_path))


def images_have_same_domain(image1, image2, tolerance=1e-5):
    # Check if the sizes and origins of the images are the same,
    # and if the spacing values are close within the given tolerance
    img1_info = get_info_from_image(image1)
    img2_info = get_info_from_image(image2)
    is_same = (
        len(img1_info.size) == len(img2_info.size)
        and all(i == j for i, j in zip(img1_info.size, img2_info.size))
        and images_have_same_spacing(image1, image2, tolerance)
        and all(
            math.isclose(i, j, rel_tol=tolerance)
            for i, j in zip(image1.GetOrigin(), image2.GetOrigin())
        )
    )
    return is_same


def images_have_same_spacing(image1, image2, tolerance=1e-5):
    # Check if the spacing values are close within the given tolerance
    is_same = all(
        math.isclose(i, j, rel_tol=tolerance)
        for i, j in zip(image1.GetSpacing(), image2.GetSpacing())
    )
    return is_same


def resample_itk_image_like(img, like_img, default_pixel_value, linear=True):
    # Create a resampler object
    ResampleFilterType = itk.ResampleImageFilter.New

    resampler = ResampleFilterType(Input=img)

    # Set the resampler parameters from like_img
    resampler.SetSize(itk.size(like_img))
    resampler.SetOutputSpacing(like_img.GetSpacing())
    resampler.SetOutputOrigin(like_img.GetOrigin())
    resampler.SetOutputDirection(like_img.GetDirection())

    # Set the default pixel value
    resampler.SetDefaultPixelValue(default_pixel_value)

    # Use the identity transform - we only resample in place
    # identity_transform = itk.IdentityTransform.New()
    # resampler.SetTransform(identity_transform)

    # Set the interpolation method to Linear if required
    if linear:
        resampler.SetInterpolator(itk.LinearInterpolateImageFunction.New(img))

    # Execute the resampling
    resampler.Update()
    resampled_img = resampler.GetOutput()

    return resampled_img


def resample_itk_image(
    image, size, spacing, default_pixel_value, linear=True, translation=None
):
    # create a temporary image
    like = create_3d_image(size, spacing, allocate=False)
    # position the image such as the centre is the same as the initial image
    info1 = get_info_from_image(like)
    center1 = info1.size / 2.0 * info1.spacing + info1.origin - info1.spacing / 2.0
    if translation is not None:
        center1 += translation
    info2 = get_info_from_image(image)
    center2 = info2.size / 2.0 * info2.spacing + info2.origin - info2.spacing / 2.0
    tr = center2 - center1
    info1.origin = tr
    like.SetOrigin(info1.origin)
    # resample
    image = resample_itk_image_like(
        image, like, default_pixel_value=default_pixel_value, linear=linear
    )
    return image



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import uproot

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test019")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.output_dir = paths.output
    sim.g4_verbose = False
    sim.visu = False
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    sim.number_of_threads = 1
    sim.random_seed = 321654
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    nm = gate.g4_units.nm
    Bq = gate.g4_units.Bq
    MeV = gate.g4_units.MeV

    #  adapt world size
    sim.world.size = [1 * m, 1 * m, 1 * m]
    sim.world.material = "G4_AIR"

    # virtual plane for phase space
    plane = sim.add_volume("Tubs", "phase_space_plane")
    plane.mother = sim.world
    plane.material = "G4_AIR"
    plane.rmin = 0
    plane.rmax = 700 * mm
    plane.dz = 1 * nm  # half height
    plane.translation = [0, 0, -100 * mm]
    plane.color = [1, 0, 0, 1]  # red

    # e- source
    source = sim.add_source("GenericSource", "Default")
    source.particle = "gamma"
    source.energy.type = "gauss"
    source.energy.mono = 1 * MeV
    source.energy.sigma_gauss = 0.5 * MeV
    source.position.type = "disc"
    source.position.radius = 20 * mm
    source.position.translation = [0, 0, 0 * mm]
    source.direction.type = "momentum"
    source.n = 66

    # add stat actor
    stats_actor = sim.add_actor("SimulationStatisticsActor", "Stats")
    stats_actor.track_types_flag = True

    # PhaseSpace Actor
    ta2 = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    ta2.attached_to = plane.name
    ta2.attributes = [
        "KineticEnergy",
        "PostPosition",
        "PrePosition",
        "PrePositionLocal",
        "ParticleName",
        "PreDirection",
        "PreDirectionLocal",
        "PostDirection",
        "TimeFromBeginOfEvent",
        "GlobalTime",
        "LocalTime",
        "EventPosition",
        "PDGCode",
    ]
    ta2.debug = False

    # run the simulation once with no particle in the phsp
    source.direction.momentum = [0, 0, 1]
    ta2.output_filename = "test019_phsp_actor_empty.root"

    # run
    sim.run(start_new_process=True)
    print(stats_actor)

    # check if empty (the root file does not exist)
    is_ok = ta2.total_number_of_entries == 0
    utility.print_test(is_ok, f"empty phase space = {ta2.total_number_of_entries}")
    print()

    # redo with the right direction
    source.direction.momentum = [0, 0, -1]
    ta2.output_filename = "test019_phsp_actor.root"
    sim.run(start_new_process=True)
    print(stats_actor)

    # check if exists and NOT empty
    hits = uproot.open(ta2.get_output_path_string())["PhaseSpace"]
    is_ok2 = source.n - 10 < hits.num_entries < source.n + 10
    utility.print_test(is_ok2, f"Number of entries = {hits.num_entries} / {source.n}")
    print()

    is_ok = is_ok and is_ok2
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from opengate.actors.coincidences import (
    coincidences_sorter,
    copy_tree_for_dump,
)
import uproot
import os
import numpy as np
from scipy.stats import wasserstein_distance
import sys


def main(dependency="test072_coinc_sorter_step1.py"):
    # test paths
    paths = utility.get_default_test_paths(__file__, output_folder="test072")

    # open root file
    root_filename = paths.output / "output_singles.root"

    # this test need output/test072/output_singles.root
    if not os.path.exists(root_filename):
        # ignore on windows
        if os.name == "nt":
            utility.test_ok(True)
            sys.exit(0)
        cmd = "python " + str(paths.current / dependency)
        r = os.system(cmd)

    print(f"Opening {root_filename} ...")
    root_file = uproot.open(root_filename)

    # consider the tree of "singles"
    singles_tree = root_file["Singles_crystal"]
    n = int(singles_tree.num_entries)
    print(f"There are {n} singles")

    # time windows
    ns = gate.g4_units.nanosecond
    ms = gate.g4_units.millisecond
    time_window = 3 * ns
    transaxial_plane = "xy"
    policy = "takeAllGoods"

    mm = gate.g4_units.mm
    minDistanceXY = (
        208.46 * mm
    )  # = sqrt(2)*147.4, where 147.4 is the mean value inside the detector block (from root histo)
    maxDistanceZ = 32 * mm
    # apply coincidences sorter
    # (chunk size can be much larger, keep a low value to check it is ok)
    coincidences = coincidences_sorter(
        singles_tree,
        time_window,
        policy,
        minDistanceXY,
        transaxial_plane,
        maxDistanceZ,
        chunk_size=1000000,
    )
    nc = len(coincidences["GlobalTime1"])
    print(f"There are {nc} coincidences for policy", policy)

    # save to file
    # WARNING root version >= 5.2.2 needed
    output_file = uproot.recreate(paths.output / f"output_minDistanceXY.root")
    output_file["coincidences"] = coincidences
    output_file["singles"] = copy_tree_for_dump(singles_tree)

    # Compare with reference output
    ref_folder = paths.output_ref

    ref_filename = ref_folder / f"minDistanceXY_Gate9.4.root"
    ref_file = uproot.open(ref_filename)
    ref_coincidences = ref_file["Coincidences"]

    nc_ref = int(ref_coincidences.num_entries)
    stat_diff = abs(nc_ref - nc) / nc_ref
    tolerance_stat = 0.05  # 5%
    print(
        f"Stat comparison {nc} (Gate10) vs. {nc_ref} (Gate9.4): {stat_diff}, tolerance {tolerance_stat}"
    )

    both_posX = np.concatenate(
        (coincidences["PostPosition_X1"], coincidences["PostPosition_X2"])
    )
    ref_both_posX = np.concatenate(
        (ref_coincidences["globalPosX1"], ref_coincidences["globalPosX2"])
    )

    both_energy = np.concatenate(
        (coincidences["TotalEnergyDeposit1"], coincidences["TotalEnergyDeposit2"])
    )
    ref_both_energy = np.concatenate(
        (ref_coincidences["energy1"], ref_coincidences["energy2"])
    )

    # Calculate Wasserstein distance for comparison
    distance_posX = wasserstein_distance(both_posX, ref_both_posX)
    tolerance_posX = 1.03
    print(f"Wasserstein distance on X : {distance_posX}, tolerence {tolerance_posX}")

    distance_energy = wasserstein_distance(both_energy, ref_both_energy)
    tolerance_energy = 0.005
    print(
        f"Wasserstein distance on energy : {distance_energy}, tolerence {tolerance_energy}"
    )

    is_ok = (
        stat_diff < tolerance_stat
        and distance_posX < tolerance_posX
        and distance_energy < tolerance_energy
    )

    utility.test_ok(is_ok)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility

if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, gate_folder="gate_test005_proton")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_engine = "MersenneTwister"

    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    MeV = gate.g4_units.MeV

    # add a simple volume
    waterbox = sim.add_volume("Box", "Waterbox")
    waterbox.size = [40 * cm, 40 * cm, 40 * cm]
    waterbox.translation = [0 * cm, 0 * cm, 25 * cm]
    waterbox.material = "G4_WATER"

    # default source for tests
    # FiXME warning ref not OK (cppSource is not exactly the same)
    source = sim.add_source("GenericSource", "Default")
    source.particle = "proton"
    source.energy.mono = 150 * MeV
    source.position.radius = 10 * mm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.n = 20000

    # add stat actor
    sim.add_actor("SimulationStatisticsActor", "Stats")

    # verbose (WARNING : sim.g4_verbose must be True !)
    sim.g4_commands_after_init.append("/tracking/verbose 0")
    # sim.g4_commands_after_init.append("/run/verbose 2")
    # sim.g4_commands_after_init.append("/event/verbose 2")
    # sim.g4_commands_after_init.append("/tracking/verbose 1")

    print(sim.source_manager.dump_sources())

    # start simulation
    sim.run()

    # get results
    stats = sim.get_actor("Stats")
    print("Simulation seed:", sim.current_random_seed)
    print(stats)

    # gate_test5_proton
    # Gate mac/main.mac
    print("-" * 80)
    stats_ref = utility.read_stat_file(paths.gate_output / "stat.txt")
    is_ok = utility.assert_stats(stats, stats_ref, tolerance=0.155)

    utility.test_ok(is_ok)



from typing import Optional

import opengate_core as g4
from ..definitions import __world_name__
from ..base import GateObject, process_cls
from ..geometry.utility import rot_np_as_g4, vec_np_as_g4
from ..exception import fatal
from .base import ActorBase
from ..decorators import requires_fatal


class DynamicGeometryActor(ActorBase, g4.GateVActor):

    def __init__(self, *args, **kwargs):
        kwargs["attached_to"] = __world_name__
        ActorBase.__init__(self, *args, **kwargs)
        self.geometry_changers = []
        self.__initcpp__()

    def __initcpp__(self):
        g4.GateVActor.__init__(self, {"name": self.name})
        self.AddActions({"BeginOfRunActionMasterThread"})

    def close(self):
        for c in self.geometry_changers:
            c.close()
        self.geometry_changers = []
        super().close()

    def to_dictionary(self):
        return_dict = super().to_dictionary()
        return_dict["geometry_changers"] = dict(
            [(v.name, v.to_dictionary()) for v in self.geometry_changers]
        )
        return return_dict

    def initialize(self):
        ActorBase.initialize(self)
        for c in self.geometry_changers:
            if c.volume_manager is None:
                c.volume_manager = self.simulation.volume_manager
            c.initialize()

    def BeginOfRunActionMasterThread(self, run_id):
        if self.simulation.dyn_geom_open_close:
            gm = g4.G4GeometryManager.GetInstance()
            # OpenGeometry (G4VPhysicalVolume *vol=0)
            gm.OpenGeometry(None)
        for c in self.geometry_changers:
            c.apply_change(run_id)
        if self.simulation.dyn_geom_open_close:
            # CloseGeometry: pOptimise=true, verbose=false, G4VPhysicalVolume *vol=0
            gm.CloseGeometry(self.simulation.dyn_geom_optimise, False, None)


def _setter_hook_attached_to(self, value):
    # try to pick up the simulation from the attached_to volume
    try:
        simulation = value.simulation
    except AttributeError:
        simulation = None
    if (
        self.simulation is not None
        and simulation is not None
        and self.simulation is not simulation
    ):
        fatal(
            f"The simulation of the changers is different from the simulation "
            f"which manages the attached_to volume. "
        )
    if simulation is not None:
        self.simulation = simulation
    try:
        return value.name
    except AttributeError:
        return value


class GeometryChanger(GateObject):

    # hints for IDE
    attached_to: Optional[str]

    user_info_defaults = {
        "attached_to": (
            None,
            {
                "doc": "The object which this changer handles, e.g. a volume.",
                "setter_hook": _setter_hook_attached_to,
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        # FIXME: update this considering that GateObject now holds a ref to simulation
        simulation = kwargs.pop(
            "simulation", None
        )  # intercept simulation from kwargs if present
        super().__init__(*args, **kwargs)

        # the user might have passed an 'attached_to' keyword argument pointing to a Volume object
        # in that case, the simulation was picked up from the volume
        # check for consistency
        if (
            self.simulation is not None
            and simulation is not None
            and self.simulation is not simulation
        ):
            fatal(
                f"The simulation passed as keyword argument is different "
                f"from the simulation which manages the attached_to volume. "
            )
        if simulation is not None:
            self.simulation = simulation

    @property
    def volume_manager(self):
        if self.simulation is not None:
            return self.simulation.volume_manager
        else:
            return None

    @property
    @requires_fatal("volume_manager")
    def attached_to_volume(self):
        return self.volume_manager.get_volume(self.attached_to)

    def initialize(self):
        # dummy implementation - nothing to do in the general case
        pass

    def apply_change(self, run_id):
        raise NotImplementedError(
            f"You are trying to call the method in the base class {type(self)}, "
            f"but it is only available in classes inheriting from it. "
        )


class VolumeImageChanger(GeometryChanger):

    # hints for IDE
    images: Optional[list]
    label_image: Optional[dict]

    user_info_defaults = {
        "images": (
            None,
            {
                "doc": "List of image names corresponding to the run timing intervals. ",
            },
        ),
        "label_image": (
            None,
            {
                "doc": "Dictionary of label images where the keys correspond to the image names "
                "stored in the user info 'images'.",
            },
        ),
    }

    def apply_change(self, run_id):
        self.attached_to_volume.update_label_image(
            self.label_image[self.images[run_id]]
        )


class VolumeTranslationChanger(GeometryChanger):

    # hints for IDE
    translations: Optional[list]
    repetition_index: int

    user_info_defaults = {
        "translations": (
            None,
            {
                "doc": "The list of translations corresponding to the run timing intervals. ",
            },
        ),
        "repetition_index": (
            0,
            {
                "doc": "The copy index of the G4PhysicalVolume to which the translations are applied. ",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.g4_translations = []
        self.g4_physical_volume = None

    def close(self):
        self.g4_translations = []
        self.g4_physical_volume = None
        super().close()

    def __getstate__(self):
        return_dict = super().__getstate__()
        return_dict["g4_translations"] = None
        return_dict["g4_physical_volume"] = None
        return return_dict

    def initialize(self):
        self.g4_physical_volume = self.attached_to_volume.get_g4_physical_volume(
            self.repetition_index
        )

        self.g4_translations = []
        for t in self.translations:
            self.g4_translations.append(vec_np_as_g4(t))

    def apply_change(self, run_id):
        self.g4_physical_volume.SetTranslation(self.g4_translations[run_id])


class VolumeRotationChanger(GeometryChanger):

    # hints for IDE
    rotations: Optional[list]
    repetition_index: int

    user_info_defaults = {
        "rotations": (
            None,
            {
                "doc": "The list of rotations corresponding to the run timing intervals. ",
            },
        ),
        "repetition_index": (
            0,
            {
                "doc": "The copy index of the G4PhysicalVolume to which the translations are applied. ",
            },
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.g4_rotations = []
        self.g4_physical_volume = None
        self.r = []

    def close(self):
        self.g4_rotations = []
        self.g4_physical_volume = None
        super().close()

    def __getstate__(self):
        return_dict = super().__getstate__()
        return_dict["g4_rotations"] = None
        return_dict["g4_physical_volume"] = None
        return return_dict

    def initialize(self):
        self.g4_physical_volume = self.attached_to_volume.get_g4_physical_volume(
            self.repetition_index
        )

        self.g4_rotations = []
        for r in self.rotations:
            g4_rot = rot_np_as_g4(r)
            g4_rot.invert()
            self.g4_rotations.append(g4_rot.rep3x3())

    def apply_change(self, run_id):
        self.g4_physical_volume.SetRotationHepRep3x3(self.g4_rotations[run_id])


process_cls(DynamicGeometryActor)
process_cls(GeometryChanger)
process_cls(VolumeImageChanger)
process_cls(VolumeTranslationChanger)
process_cls(VolumeRotationChanger)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itk
from scipy.spatial.transform import Rotation
import opengate as gate
from opengate.tests import utility
from opengate.contrib.beamlines.ionbeamline import BeamlineModel
import numpy as np

from opengate.tests.utility import print_test

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "gate_test044_pbs", output_folder="test059"
    )
    output_path = paths.output
    ref_path = paths.output_ref

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    sim.g4_verbose_level = 1
    sim.visu = False
    sim.random_seed = 12365478910
    sim.random_engine = "MersenneTwister"
    sim.output_dir = output_path
    sim.number_of_threads = 2

    # units
    km = gate.g4_units.km
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg
    rad = gate.g4_units.rad

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]

    # nozzle box
    box = sim.add_volume("Box", "box")
    box.size = [500 * mm, 500 * mm, 1000 * mm]
    box.translation = [1148 * mm, 0.0, 0.0]
    box.rotation = Rotation.from_euler("y", -90, degrees=True).as_matrix()
    box.material = "Vacuum"
    box.color = [0, 0, 1, 1]

    # nozzle WET
    nozzle = sim.add_volume("Box", "nozzle")
    nozzle.mother = box.name
    nozzle.size = [500 * mm, 500 * mm, 2 * mm]
    nozzle.material = "G4_WATER"

    # target
    phantom = sim.add_volume("Box", "phantom")
    phantom.size = [500 * mm, 500 * mm, 400 * mm]
    phantom.rotation = Rotation.from_euler("y", 90, degrees=True).as_matrix()
    phantom.translation = [-200.0, 0.0, 0]
    phantom.material = "G4_WATER"
    phantom.color = [0, 0, 1, 1]

    # roos chamber
    peak_finder = sim.add_volume("Tubs", "peak_finder")
    peak_finder.mother = phantom.name
    peak_finder.material = "G4_WATER"
    peak_finder.rmax = 40.3
    peak_finder.rmin = 0
    peak_finder.dz = 200
    peak_finder.color = [1, 0, 1, 1]

    # physics
    sim.physics_manager.physics_list_name = "FTFP_INCLXX_EMZ"
    sim.physics_manager.set_production_cut("world", "all", 1000 * km)

    # add dose actor
    dose_actor = sim.add_actor("DoseActor", "doseInXYZ")
    dose_actor.output_filename = "dose_peak_finder.mhd"
    dose_actor.attached_to = peak_finder.name
    dose_actor.size = [1, 1, 8000]
    dose_actor.spacing = [80.6, 80.6, 0.05]
    dose_actor.hit_type = "random"
    dose_actor.dose.active = True

    # ---------- DEFINE BEAMLINE MODEL -------------
    IR2HBL = BeamlineModel()
    IR2HBL.name = None
    IR2HBL.radiation_types = "ion 6 12"
    # Nozzle entrance to Isocenter distance
    IR2HBL.distance_nozzle_iso = 1300.00  # 1648 * mm#1300 * mm
    # SMX to Isocenter distance
    IR2HBL.distance_stearmag_to_isocenter_x = 2000.00
    # SMY to Isocenter distance
    IR2HBL.distance_stearmag_to_isocenter_y = 2000.00
    # polinomial coefficients
    IR2HBL.energy_mean_coeffs = [12.0, -9.54]
    IR2HBL.energy_spread_coeffs = [0.00048, 5.2532]
    IR2HBL.sigma_x_coeffs = [2.33]
    IR2HBL.theta_x_coeffs = [0.00029]
    IR2HBL.epsilon_x_coeffs = [0.00078]
    IR2HBL.sigma_y_coeffs = [1.96]
    IR2HBL.theta_y_coeffs = [0.00079]
    IR2HBL.epsilon_y_coeffs = [0.0024]

    # --------START PENCIL BEAM SCANNING----------
    # NOTE: HBL means that the beam is coming from -x (90 degree rot around y)

    tps = sim.add_source("TreatmentPlanPBSource", "TPSource")
    tps.n = 20000
    tps.beam_model = IR2HBL
    tps.plan_path = ref_path / "PlanCentralSpot_1440MeV.txt"
    tps.beam_nr = 1
    tps.particle = "ion 6 12"

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "Stats")
    # s.track_types_flag = True

    # start simulation
    sim.run()

    # -------------END SCANNING-------------
    # print results at the end
    print(stats)

    # ------ TESTS -------##
    dose_path = output_path / sim.get_actor("doseInXYZ").dose.get_output_path()

    # RANGE

    # read output and ref
    print("Compare ", dose_path)
    img_mhd_out = itk.imread(dose_path)
    # data = np.flip(itk.GetArrayViewFromImage(img_mhd_out), axis=0)
    data = itk.GetArrayViewFromImage(img_mhd_out)
    spacing = np.array(img_mhd_out.GetSpacing())
    print(data.shape, spacing)

    # Range 80
    range80_gate9_E120MeV = 367.06
    range_opengate = utility.get_range_from_image(data, data.shape, spacing, axis="z")

    thresh = 2.0 * mm
    ok = True
    print(f"range_opengate = {range_opengate}")
    print(f"range80_gate9_E120MeV = {range80_gate9_E120MeV}")
    if abs(range_opengate - range80_gate9_E120MeV) > thresh:
        ok = False
    print_test(
        ok, f"Compare ranges {range_opengate} and {range80_gate9_E120MeV} >? {thresh}"
    )

    utility.test_ok(ok)



import numpy as np
import pathlib
import SimpleITK as sitk
import itk
from opengate.geometry.utility import (
    translate_point_to_volume,
    vec_g4_as_np,
)
from opengate.actors.digitizers import *
import sys


def add_fake_table(sim, name="table"):
    """
    Add a patient table (fake)
    """

    # unit
    mm = g4_units.mm
    cm = g4_units.cm
    cm3 = g4_units.cm3
    deg = g4_units.deg
    gcm3 = g4_units.g / cm3

    # colors
    red = [1, 0.7, 0.7, 0.8]
    white = [1, 1, 1, 1]

    sim.volume_manager.material_database.add_material_weights(
        f"CarbonFiber", ["C"], [1], 1.78 * gcm3
    )

    # main bed
    table = sim.add_volume("Tubs", f"{name}_table")
    table.mother = "world"
    table.rmax = 439 * mm
    table.rmin = 406 * mm
    table.dz = 200 * cm / 2.0
    table.sphi = 0 * deg
    table.dphi = 70 * deg
    table.translation = [0, -25 * cm, 0]
    table.rotation = Rotation.from_euler("z", 55, degrees=True).as_matrix()
    table.material = "CarbonFiber"
    table.color = white

    # interior of the table
    tablein = sim.add_volume("Tubs", f"{name}_tablein")
    tablein.mother = table.name
    tablein.rmax = 436.5 * mm
    tablein.rmin = 408.5 * mm
    tablein.dz = 200 * cm / 2.0
    tablein.sphi = 0 * deg
    tablein.dphi = 69 * deg
    tablein.translation = [0, 0, 0]
    tablein.rotation = Rotation.from_euler("z", 0.5, degrees=True).as_matrix()
    tablein.material = "G4_AIR"
    tablein.color = red

    return table


def add_fake_table_OLD(sim, name="table"):
    """
    Add a patient table (fake)
    """

    # unit
    mm = g4_units.mm
    cm = g4_units.cm
    cm3 = g4_units.cm3
    deg = g4_units.deg
    gcm3 = g4_units.g / cm3

    # colors
    red = [1, 0.7, 0.7, 0.8]
    white = [1, 1, 1, 1]

    sim.volume_manager.material_database.add_material_weights(
        f"CarbonFiber", ["C"], [1], 1.78 * gcm3
    )

    # main bed
    table = sim.add_volume("Tubs", f"{name}_table")
    table.mother = "world"
    table.rmax = 439 * mm
    table.rmin = 406 * mm
    table.dz = 200 * cm / 2.0
    table.sphi = 0 * deg
    table.dphi = 70 * deg
    table.translation = [0, 25 * cm, 0]
    table.rotation = Rotation.from_euler("z", -125, degrees=True).as_matrix()
    table.material = "CarbonFiber"
    table.color = white

    # interior of the table
    tablein = sim.add_volume("Tubs", f"{name}_tablein")
    tablein.mother = table.name
    tablein.rmax = 436.5 * mm
    tablein.rmin = 408.5 * mm
    tablein.dz = 200 * cm / 2.0
    tablein.sphi = 0 * deg
    tablein.dphi = 69 * deg
    tablein.translation = [0, 0, 0]
    tablein.rotation = Rotation.from_euler("z", 0.5, degrees=True).as_matrix()
    tablein.material = "G4_AIR"
    tablein.color = red

    return table


def get_volume_position_in_head(sim, spect_name, vol_name, pos="max", axis=2):
    vol = sim.volume_manager.volumes[f"{spect_name}_{vol_name}"]
    pMin, pMax = vol.bounding_limits
    x = pMax
    if pos == "min":
        x = pMin
    if pos == "center":
        x = pMin + (pMax - pMin) / 2.0
    x = vec_g4_as_np(x)
    x = translate_point_to_volume(sim, vol, spect_name, x)
    return x[axis]


def extract_energy_window_itk(
    image: itk.Image,
    energy_window: int = "all",
    nb_of_energy_windows: int = 3,
    nb_of_gantries: int = 60,
) -> itk.Image:
    """
    Extracts a given energy window from a 3D image and returns a reshaped image.

    3rd dim is number_of_energy_window x number_of_gantries

    Args:
        image (itk.Image): The input 3D image of shape (e.g. 128 x 128 x 3 x 60).
        energy_window : which energy windows to extract (int or 'all').
        nb_of_energy_windows (int): the number of energy windows
        nb_of_gantries (int): The number of gantry positions (default: 60).

    Returns:
        itk.Image: The extracted and reshaped image of shape (128 x 128 x 60).
    """

    # Validate input dimensions
    width, height, depth = image.GetSize()
    if depth != nb_of_gantries * nb_of_energy_windows:
        raise ValueError(
            f"Input image depth {depth} does not match expected "
            f"dimensions = {nb_of_gantries} x {nb_of_energy_windows}."
        )

    # Extract the relevant slices
    # (warning numpy is z,x,y, while itk is x,y,z)
    arr = itk.GetArrayViewFromImage(image)
    sub_image_array = arr[energy_window::nb_of_energy_windows, :, :]

    # Convert the numpy array back to a SimpleITK image
    sub_image = itk.GetImageFromArray(sub_image_array)

    # Preserve the original image metadata
    sub_image.SetSpacing(image.GetSpacing())
    sub_image.SetOrigin(image.GetOrigin())
    sub_image.SetDirection(image.GetDirection())

    return sub_image


def extract_energy_window_from_projection_actors(
    projections,
    energy_window: int = "all",
    nb_of_energy_windows: int = 3,
    nb_of_gantries: int = 60,
    output_filenames=None,
):
    auto_output_filenames = output_filenames is None
    if auto_output_filenames:
        output_filenames = []
    i = 0
    for proj in projections:
        filename = proj.get_output_path()
        img = itk.imread(filename)
        out = extract_energy_window_itk(
            img,
            energy_window=energy_window,
            nb_of_energy_windows=nb_of_energy_windows,
            nb_of_gantries=nb_of_gantries,
        )
        if auto_output_filenames:
            extension = pathlib.Path(filename).suffix
            output_filename = str(filename).replace(
                extension, f"_ene_{energy_window}{extension}"
            )
            output_filenames.append(output_filename)
        else:
            output_filename = output_filenames[i]
        itk.imwrite(out, output_filename)
        i += 1
    return output_filenames


def merge_several_heads_projections(filenames):
    output_arr = None
    img = None
    for filename in filenames:
        img = itk.imread(filename)
        arr = itk.GetArrayFromImage(img)
        if output_arr is None:
            output_arr = arr
        else:
            output_arr = np.concatenate((output_arr, arr), axis=0)

    output_img = itk.GetImageFromArray(output_arr)
    output_img.CopyInformation(img)
    return output_img


def read_projections_as_sinograms(filenames, nb_of_gantry_angles):
    """
    Reads projection files from a specified folder, processes them into sinograms per
    energy window, and ensures consistency in image metadata across all projections.

    Args:
        filenames : List of projection filenames to read.
        nb_of_gantry_angles (int): Number of gantry angles in the projections.

    Returns:
        list[sitk.Image]: List of SimpleITK Image objects containing the sinograms per
        energy window.
    """
    # init variables
    sinograms_per_energy_window = None
    nb_of_energy_windows = None
    projection_size = None
    projection_origin = None
    projection_spacing = None

    # read all projection files (one per head)
    for f in filenames:
        img = sitk.ReadImage(f)

        if nb_of_energy_windows is None:
            nb_of_energy_windows = int(img.GetSize()[2] / nb_of_gantry_angles)
            projection_size = img.GetSize()
            projection_origin = img.GetOrigin()
            projection_spacing = img.GetSpacing()
            sinograms_per_energy_window = [None] * nb_of_energy_windows

        if nb_of_energy_windows < 1:
            raise ValueError(
                f"nb_of_energy_windows={nb_of_energy_windows} is invalid ; "
                f"image size is {img.GetSize()} and "
                f"nb of angles is {nb_of_gantry_angles}"
            )

        # check that size and origin are the same for all images
        if img.GetSize() != projection_size:
            raise ValueError(
                f"Projections in {f} have different size than in {filenames[0]}"
            )
        if img.GetOrigin() != projection_origin:
            raise ValueError(
                f"Projections in {f} have different origin than in {filenames[0]}"
            )
        if img.GetSpacing() != projection_spacing:
            raise ValueError(
                f"Projections in {f} have different spacing than in {filenames[0]}"
            )

        # convert to a numpy array
        arr = sitk.GetArrayViewFromImage(img)

        # concatenate projections for the different heads and for each energy window
        for ene in range(nb_of_energy_windows):
            # this is important to make a copy here!
            # Otherwise, the concatenate operation may fail later
            a = arr[ene::nb_of_energy_windows, :, :].copy()
            if sinograms_per_energy_window[ene] is None:
                sinograms_per_energy_window[ene] = a
            else:
                sinograms_per_energy_window[ene] = np.concatenate(
                    (sinograms_per_energy_window[ene], a), axis=0
                )

    # build sitk image from np arrays (to keep spacing, origin information)
    sinograms = []
    for ene in range(nb_of_energy_windows):
        img = sitk.GetImageFromArray(sinograms_per_energy_window[ene])
        img.SetSpacing(projection_spacing)
        img.SetOrigin(projection_origin)
        img.SetDirection(img.GetDirection())
        sinograms.append(img)
    return sinograms


def poisson_rel_uncertainty(np_image):
    """
    Uncertainty for Poisson counts is sqrt(counts) = standard deviation
    Relative uncertainty is sqrt(counts)/counts
    """
    relative_uncertainty = np_image
    uncertainty = np.sqrt(np_image)
    relative_uncertainty = np.divide(
        uncertainty,
        relative_uncertainty,
        out=np.zeros_like(relative_uncertainty),
        where=relative_uncertainty != 0,
    )
    return relative_uncertainty


def poisson_rel_uncertainty_from_files(input_filename, output_rel_uncert=None):
    """ """
    img = sitk.ReadImage(input_filename)
    relative_uncertainty = sitk.GetArrayFromImage(img)
    relative_uncertainty = poisson_rel_uncertainty(relative_uncertainty)
    if output_rel_uncert is not None:
        uncert = sitk.GetImageFromArray(relative_uncertainty)
        uncert.CopyInformation(img)
        sitk.WriteImage(uncert, output_rel_uncert)
    return relative_uncertainty


def batch_rel_uncertainty(np_images):
    """
    Computes the mean and the relative uncertainty of the given images as np arrays.
    Parameters:
        np_images (list of np.ndarray): List of Numpy arrays from images.
    Returns:
        np.ndarray: The mean image.
        np.ndarray: The relative uncertainty image.
    """
    mean = None
    nb_batch = len(np_images)

    # Compute mean
    for m in np_images:
        if mean is None:
            mean = m.copy()
        else:
            mean += m
    mean /= nb_batch

    # Compute standard deviation (variance)
    squared = None
    for m in np_images:
        diff_squared = np.power(m - mean, 2)
        if squared is None:
            squared = diff_squared
        else:
            squared += diff_squared

    # Compute uncertainty
    uncert = np.sqrt(np.divide(squared, nb_batch * (nb_batch - 1)))

    # Compute relative uncertainty in %
    uncert = np.divide(uncert, mean, out=np.zeros_like(mean), where=mean != 0)

    return mean, uncert


def batch_rel_uncertainty_from_files(
    image_filenames, mean_filename=None, rel_uncert_filename=None
):
    """
    Wrapper function that reads images from file, computes relative uncertainty, and writes results to files.
    Parameters:
        image_filenames (list of str): List of image file paths.
        mean_filename (str): Path to save the mean image (optional).
        rel_uncert_filename (str): Path to save the relative uncertainty image (optional).
    Returns:
        np.ndarray: The mean image.
        np.ndarray: The relative uncertainty image.
    """
    # Read images into Numpy arrays
    np_images = [sitk.GetArrayFromImage(sitk.ReadImage(f)) for f in image_filenames]

    # Compute mean and relative uncertainty
    mean, uncert = batch_rel_uncertainty(np_images)

    # Write results back to files if filenames are provided
    if mean_filename is not None:
        img = sitk.GetImageFromArray(mean)
        img.CopyInformation(sitk.ReadImage(image_filenames[0]))
        sitk.WriteImage(img, str(mean_filename))
    if rel_uncert_filename is not None:
        img = sitk.GetImageFromArray(uncert)
        img.CopyInformation(sitk.ReadImage(image_filenames[0]))
        sitk.WriteImage(img, str(rel_uncert_filename))

    return mean, uncert


def compute_efficiency_from_files(filename, duration):
    img_ref = sitk.ReadImage(str(filename))
    np_uncert = sitk.GetArrayFromImage(img_ref)
    return compute_efficiency(np_uncert, duration)


def compute_efficiency(np_uncert, duration):
    ones = np.ones_like(np_uncert)
    eff = np.divide(
        ones,
        (np_uncert * np_uncert) * duration,
        out=np.zeros_like(np_uncert),
        where=np_uncert != 0,
    )
    return eff


def history_rel_uncertainty(np_img, np_img_squared, n):
    mean = np_img / n
    squared = np_img_squared / n
    variance = (squared - np.power(mean, 2)) / (n - 1)
    uncertainty = np.divide(
        np.sqrt(variance),
        mean,
        out=np.zeros_like(variance),
        where=mean != 0,
    )
    return uncertainty


def history_rel_uncertainty_from_files(
    img_filename,
    img_squared_filename,
    n,
    output_filename=None,
):
    # primary
    img = sitk.ReadImage(img_filename)
    img_squared = sitk.ReadImage(img_squared_filename)
    mean = sitk.GetArrayFromImage(img)
    squared = sitk.GetArrayFromImage(img_squared)

    # compute
    uncert = history_rel_uncertainty(mean, squared, n)

    if output_filename is not None:
        img_uncert = sitk.GetImageFromArray(uncert)
        img_uncert.CopyInformation(img)
        sitk.WriteImage(img_uncert, output_filename)

    return uncert, mean, squared


def history_ff_combined_rel_uncertainty(
    vprim, vprim_squared, vscatter, vscatter_squared, n_prim, n_scatter
):

    # means for one event
    if vprim is not None:
        prim = vprim / n_prim
        prim_squared = vprim_squared / n_prim
        prim_var = (prim_squared - np.power(prim, 2)) / (n_prim - 1)
        mean = prim
        variance = prim_var
    if vscatter is not None:
        scatter = vscatter / n_scatter
        scatter_squared = vscatter_squared / n_scatter
        scatter_var = (scatter_squared - np.power(scatter, 2)) / (n_scatter - 1)
        mean = scatter
        variance = scatter_var

    if vprim is not None and vscatter is not None:
        mean = prim + scatter
        variance = prim_var + scatter_var

    uncert = np.divide(
        np.sqrt(variance),
        mean,
        out=np.zeros_like(variance),
        where=mean != 0,
    )

    # rescale the mean for the final results
    mean = mean * n_prim

    return uncert, mean


def batch_ff_combined_rel_uncertainty(
    prim_mean, prim_uncert, scatter_mean, scatter_uncert, n_prim, n_scatter
):
    # combine mean
    r = n_prim / n_scatter
    mean = prim_mean + scatter_mean * r

    # combine uncertainties
    prim_var = np.power(prim_uncert * prim_mean, 2)
    sc_var = np.power(scatter_uncert * scatter_mean, 2) * np.power(r, 2)
    uncert = np.sqrt(prim_var + sc_var)
    uncert = np.divide(
        uncert,
        mean,
        out=np.zeros_like(uncert),
        where=mean != 0,
    )

    return uncert, mean


def get_default_energy_windows(radionuclide_name, spectrum_channel=False):
    n = radionuclide_name.lower()
    keV = g4_units.keV
    channels = []

    if "177lu" in n or "lu177" in n:
        channels = [
            {"name": f"spectrum", "min": 3 * keV, "max": 515 * keV},
            {"name": f"scatter1", "min": 84.75 * keV, "max": 101.7 * keV},
            {"name": f"peak113", "min": 101.7 * keV, "max": 124.3 * keV},
            {"name": f"scatter2", "min": 124.3 * keV, "max": 141.25 * keV},
            {"name": f"scatter3", "min": 145.6 * keV, "max": 187.2 * keV},
            {"name": f"peak208", "min": 187.2 * keV, "max": 228.8 * keV},
            {"name": f"scatter4", "min": 228.8 * keV, "max": 270.4 * keV},
        ]
    if "tc99m" in n:
        channels = [
            {"name": f"spectrum", "min": 3 * keV, "max": 160 * keV},
            {"name": f"scatter", "min": 108.58 * keV, "max": 129.59 * keV},
            {"name": f"peak140", "min": 129.59 * keV, "max": 150.61 * keV},
        ]
    if "in111" in n or "111in" in n:
        # 15% around the peaks
        channels = [
            {"name": "spectrum_full", "min": 3.0, "max": 515.0},
            {"name": "scatter_171_low", "min": 138.4525, "max": 158.4525},
            {"name": "peak_171", "min": 158.4525, "max": 184.1475},
            {"name": "scatter_171_high", "min": 184.1475, "max": 204.1475},
            {"name": "scatter_245_low", "min": 206.995, "max": 226.995},
            {"name": "peak_245", "min": 226.995, "max": 263.805},
            {"name": "scatter_245_high", "min": 263.805, "max": 283.805},
        ]
    if not spectrum_channel:
        channels.pop(0)
    if len(channels) == 0:
        raise ValueError(f"No default energy windows for {radionuclide_name}")
    return channels


def get_mu_from_xraylib(material_symbol, energy):
    """
    Retrieves the linear attenuation coefficient (mu) for a given material and energy.
    Uses the xraylib library to get data from standard physics databases.

    Args:
        material_symbol (str): The chemical symbol of the material (e.g. 'Pb', 'W').
        energy (float): The photon energy (will be used in keV)

    Returns:
        float: The linear attenuation coefficient in cm^-1, or None if the material is unknown.
    """
    try:
        import xraylib
    except ImportError:
        print(
            "Error: The 'xraylib' library is required for dynamic calculations of attenuation coefficients."
        )
        print("Please install it using: pip install xraylib")
        sys.exit(1)

    try:
        # Get atomic number and density for the material from xraylib's database
        atomic_number = xraylib.SymbolToAtomicNumber(material_symbol)
        density = xraylib.ElementDensity(atomic_number)
        # print(f'energy = {energy/ g4_units.keV} keV, material = "{material_symbol}"')

        # Get the total mass attenuation coefficient (cm^2/g).
        # Xraylib expects energy in keV for this function.
        mass_attenuation_coeff = xraylib.CS_Total(atomic_number, energy / g4_units.keV)

        # Calculate linear attenuation coefficient: mu = (mu/rho) * rho
        linear_attenuation_coeff = mass_attenuation_coeff * density

        return linear_attenuation_coeff

    except ValueError:
        print(
            f"Error: Material symbol '{material_symbol}' not found in xraylib database."
        )
        return None


def calculate_acceptance_angle(
    hole_diameter, collimator_length, linear_attenuation_coeff_cm
):
    """
    Calculates the effective length and acceptance angle of a collimator.

    Args:
        hole_diameter (float): The diameter of the collimator hole in mm.
        collimator_length (float): The physical length of the collimator in mm.
        linear_attenuation_coeff_cm (float): The linear attenuation coefficient
                                             of the septal material in cm^-1.

    Returns:
        tuple: A tuple containing (effective_length_mm, acceptance_angle_degrees).
    """

    mm = g4_units.mm

    # Convert mu to mm^-1
    mu_mm = linear_attenuation_coeff_cm / 10.0

    # Calculate effective length: Leff = L - 2/mu
    effective_length_mm = collimator_length / mm - (2 / mu_mm)

    # Calculate acceptance angle: theta = arctan(d / Leff)
    acceptance_angle_rad = np.arctan((hole_diameter / mm) / effective_length_mm)

    # Convert to degrees
    acceptance_angle_degrees = np.rad2deg(acceptance_angle_rad)

    return effective_length_mm, acceptance_angle_degrees


def calculate_max_penetration_angle_OLD(
    septal_thickness_mm, linear_attenuation_coeff_cm, prob_threshold=0.01
):
    """
    Calculates the maximum acceptance angle based on septal penetration
    probability.

    This angle represents the point at which the probability of a photon
    passing through the shortest path in a septum drops to a given threshold.

    Args:
        septal_thickness_mm (float): The thickness of the collimator septa in mm.
        linear_attenuation_coeff_cm (float): The linear attenuation coefficient
                                             of the septal material in cm^-1.
        prob_threshold (float, optional): The transmission probability threshold.
                                          Defaults to 0.01 (1%).

    Returns:
        float: The maximum acceptance angle in degrees.
    """
    # Ensure the probability threshold is valid
    if not 0 < prob_threshold < 1:
        raise ValueError("Probability threshold must be between 0 and 1.")

    # Convert the linear attenuation coefficient from cm^-1 to mm^-1
    mu_mm = linear_attenuation_coeff_cm / 10.0

    # Calculate the argument for arcsin: (-mu * t) / ln(P_th)
    # The numerator is negative, and ln(P_th) is also negative, so the
    # result is positive.
    arcsin_arg = (-mu_mm * septal_thickness_mm) / np.log(prob_threshold)

    # Check if the argument is valid for arcsin (it must be between -1 and 1)
    if arcsin_arg > 1:
        # This can happen if the septa are very thick or mu is very high,
        # making penetration even at 90 degrees less likely than the threshold.
        # In this case, the effective max angle is 90 degrees.
        return 90.0

    # Calculate the angle in radians
    angle_rad = np.arcsin(arcsin_arg)

    # Convert the angle to degrees
    angle_deg = np.rad2deg(angle_rad)

    return angle_deg


def calculate_max_penetration_angle(
    hole_diameter_mm: float,
    collimator_length_mm: float,
    septal_thickness_mm: float,
    linear_attenuation_coeff_cm: float,
    strictness_s: float,
    accurate_cutoff: float = 0.001,
) -> float:
    """
    Calculates the max acceptance angle using a "Strictness" parameter.

    The model interpolates between a purely geometric angle (S=1) and a
    physically accurate angle that includes penetration (S=0).

    Args:
        hole_diameter_mm (float): Diameter of the collimator hole in mm.
        collimator_length_mm (float): Physical length of the collimator in mm.
        septal_thickness_mm (float): Thickness of the collimator septa in mm.
        linear_attenuation_coeff_cm (float): Linear attenuation coefficient
                                             of the septal material in cm^-1.
        strictness_s (float): The strictness parameter, from 0 to 1.
                              S=1 is maximally strict (geometric only).
                              S=0 is minimally strict (fully accurate).
        accurate_cutoff (float, optional): The internal probability cutoff used
                                           to define the 'fully accurate' angle.
                                           Defaults to 0.001 (0.1%).

    Returns:
        float: The maximum acceptance angle in degrees.
    """
    if not 0 <= strictness_s <= 1:
        raise ValueError("Strictness (S) must be between 0 and 1.")

    # --- 1. Calculate the Geometric Angle (S=1 case) ---
    theta_geom = np.rad2deg(np.arctan(hole_diameter_mm / collimator_length_mm))

    # --- 2. Calculate the "Fully Accurate" Angle (S=0 case) ---
    # This is the angle where transmission probability drops to the low cutoff.
    # It represents the widest plausible angle including penetration.
    mu_mm = linear_attenuation_coeff_cm / 10.0
    try:
        # From the physical model: sin(theta) = -mu*t / ln(cutoff)
        arcsin_arg = (-mu_mm * septal_thickness_mm) / np.log(accurate_cutoff)

        if 0 < arcsin_arg < 1:
            theta_accurate = np.rad2deg(np.arcsin(arcsin_arg))
        else:
            # If arg is invalid (e.g., > 1), penetration is essentially
            # impossible. The most accurate model is the geometric one.
            theta_accurate = theta_geom

    except (ValueError, ZeroDivisionError):
        # Handle invalid log() input or other math errors
        theta_accurate = theta_geom

    # --- 3. Interpolate using the Strictness parameter S ---
    # S=1 gives theta_geom, S=0 gives theta_accurate.
    theta_max = strictness_s * theta_geom + (1 - strictness_s) * theta_accurate

    return theta_max


def calculate_theta_max_angle(
    hole_diameter_mm: float,
    collimator_length_mm: float,
    septal_thickness_mm: float,
    linear_attenuation_coeff_cm: float,
) -> float:

    # Calculate the Geometric Angle
    theta_geom = np.rad2deg(np.arctan(hole_diameter_mm / collimator_length_mm))

    # Calculate the effective length, according to mu
    Leff = collimator_length_mm - 2 / linear_attenuation_coeff_cm

    # Calculate the effective angle, according to mu
    theta_acc = np.rad2deg(np.arctan(hole_diameter_mm / Leff))

    # Calculate the crossover max angle
    theta_cross = np.rad2deg(
        np.arctan((hole_diameter_mm + septal_thickness_mm) / collimator_length_mm)
    )

    # print
    print(f"Geometric angle: {theta_geom:.2f} deg")
    print(f"Effective length: {Leff:.2f} mm vs {collimator_length_mm:.2f} mm")
    print(f"Effective angle: {theta_acc:.2f} deg")
    print(f"Crossover angle: {theta_cross:.2f} deg")

    return theta_acc



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import itk

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, "", output_folder="test089_geometries"
    )

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    mm = gate.g4_units.mm
    um = gate.g4_units.um
    MeV = gate.g4_units.MeV
    Bq = gate.g4_units.Bq
    nm = gate.g4_units.nm
    deg = gate.g4_units.deg

    # colors
    colors = {
        "cyan": [0, 1, 1, 1],
        "grey": [0.7, 0.7, 0.7, 1],
        "yellow": [1, 1, 0, 1],
        "green": [0, 1, 0, 1],
        "red": [1, 0, 0, 1],
        "pink": [1, 0.75, 0.79, 1],
        "orange": [1, 0.5, 0, 1],
        "blue": [0, 0, 1, 1],
        "black": [0, 0, 0, 1],
        "white": [1, 1, 1, 1],
    }

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.output_dir = paths.output

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.g4_verbose_level = 1
    ui.visu = False
    ui.visu_type = "vrml"
    ui.random_seed = "auto"
    ui.number_of_threads = 1

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # change world size
    world = sim.world
    world.size = [0.5 * m, 0.5 * m, 0.5 * m]

    # 2 cm side water box
    vol1 = sim.add_volume("Box", "vol1name")
    vol1.size = [2 * cm, 2 * cm, 2 * cm]
    vol1.translation = [0 * cm, 0 * cm, 0 * cm]
    vol1.material = "G4_WATER"
    vol1.color = colors["cyan"]

    # Half water shell centered at (0,5,0) along X, Y and Z
    vol2 = sim.add_volume("Sphere", "vol2name")
    vol2.rmin = 1 * cm
    vol2.rmax = 2 * cm
    vol2.stheta = 0 * deg
    vol2.dtheta = 90 * deg
    vol2.sphi = 0 * deg
    vol2.dphi = 360 * deg
    vol2.translation = [0 * cm, 5 * cm, 0 * cm]
    vol2.material = "G4_TISSUE_SOFT_ICRP"
    vol2.color = colors["yellow"]

    # Half 6 cm height concrete cylinder centered at (0, 10,0) cm in X, Y and Z
    vol3 = sim.add_volume("Tubs", "vol3name")
    vol3.rmin = 0.5 * cm
    vol3.rmax = 1 * cm
    vol3.dz = 3 * cm
    vol3.sphi = 0 * deg
    vol3.dphi = 180 * deg
    vol3.translation = [0 * cm, 10 * cm, 0 * cm]
    vol3.material = "G4_BRAIN_ICRP"
    vol3.color = colors["green"]

    # Trapezoid
    vol4 = sim.add_volume("Trap", "vol4name")
    vol4.dz = 60 * mm
    vol4.theta = 20 * deg
    vol4.phi = 5 * deg
    vol4.dy1 = 40 * mm
    vol4.dx1 = 30 * mm
    vol4.dx2 = 40 * mm
    vol4.alp1 = 10 * deg
    vol4.dy2 = 16 * mm
    vol4.dx3 = 10 * mm
    vol4.dx4 = 14 * mm
    vol4.alp2 = 10 * deg
    vol4.translation = [0 * cm, 0 * cm, -10 * cm]
    vol4.material = "G4_CONCRETE"
    vol4.color = colors["red"]

    # Trapezoid
    vol5 = sim.add_volume("Trd", "vol5name")
    vol5.dx1 = 30 * mm
    vol5.dx2 = 40 * mm
    vol5.dy1 = 10 * mm
    vol5.dy2 = 16 * mm
    vol5.dz = 10 * mm
    vol5.translation = [0 * cm, 7 * cm, -10 * cm]
    vol5.material = "G4_KAPTON"
    vol5.color = colors["pink"]

    # Polyhedra
    vol6 = sim.add_volume("Polyhedra", "vol6name")
    vol6.phi_start = 0 * deg
    vol6.phi_total = 360 * deg
    vol6.num_side = 6
    vol6.num_zplanes = 2
    vol6.zplane = [-2.5 * mm, 2.5 * mm]
    vol6.radius_inner = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    vol6.radius_outer = [0.15 * mm] * 6
    vol6.translation = [0 * cm, 13 * cm, -10 * cm]
    vol6.material = "G4_LUNG_ICRP"
    vol6.color = colors["orange"]

    # Cons
    vol7 = sim.add_volume("Cons", "vol7name")
    vol7.rmin1 = 0 * mm
    vol7.rmax1 = 2 * cm
    vol7.rmin2 = 2 * cm
    vol7.rmax2 = 4 * cm
    vol7.dz = 5 * cm
    vol7.sphi = 0 * deg
    vol7.dphi = 270 * deg
    vol7.translation = [0 * cm, 18 * cm, -10 * cm]
    vol7.material = "G4_PARAFFIN"
    vol7.color = colors["blue"]

    # Hexagon
    vol8 = sim.add_volume("Hexagon", "vol8name")
    vol8.height = 3 * cm
    vol8.radius = 2 * cm
    vol8.translation = [0 * cm, 18 * cm, 10 * cm]
    vol8.material = "Adipose"
    vol8.color = colors["black"]

    # Ellipsoid
    vol9 = sim.add_volume("Ellipsoid", "vol9name")
    vol9.xSemiAxis = 4 * cm
    vol9.ySemiAxis = 2 * cm
    vol9.zSemiAxis = 6 * cm
    vol9.zBottomCut = 1 * cm
    vol9.zTopCut = 4 * cm
    vol9.translation = [0 * cm, 13 * cm, 10 * cm]
    vol9.material = "Muscle"
    vol9.color = colors["white"]

    # Image Volume example in test009_voxels.py
    # Boolean Volume example in test016_bool_volumes.py
    # Repeatable Volume example in test017_repeatable.py
    # Tesselated Volume example in test067_tesselated_volume.py

    labels, image = sim.voxelize_geometry(extent="Auto", spacing=(2, 2, 2), margin=1)
    itk.imwrite(image, paths.output / "test089_geometries.mhd")

    # physics

    # source

    # dose actor

    # add stat actor

    # start simulation
    sim.run()

    # compare with ref
    ok = utility.assert_images(
        paths.output_ref / "test089_geometries.mhd",
        paths.output / "test089_geometries.mhd",
        tolerance=1e-5,
        sum_tolerance=1e-5,
    )
    # print results at the end
    utility.test_ok(ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.linacs.elektaversa as versa
import scipy.optimize
from opengate.tests import utility
import numpy as np
import uproot


def is_ok_test019(rootfile, x_field, y_field, tol=0.15):
    x = rootfile["PrePosition_X"][rootfile["ParticleName"] == "alpha"]
    y = rootfile["PrePosition_Y"][rootfile["ParticleName"] == "alpha"]

    hist_x_pos = np.histogram(
        x, bins=50, density=True, range=[-1.5 * x_field / 2, 1.5 * x_field / 2]
    )
    x_hist_x_pos = hist_x_pos[1][:-1] + 0.5 * (hist_x_pos[1][1] - hist_x_pos[1][0])
    median_hist_x_pos = np.median(hist_x_pos[0][hist_x_pos[0] > 0])
    y_hist_x_pos = hist_x_pos[0]
    fwhm_x_pos = (
        -x_hist_x_pos[np.where(hist_x_pos[0] > median_hist_x_pos / 3)[0]][0]
        + x_hist_x_pos[np.where(hist_x_pos[0] > median_hist_x_pos / 3)[0]][-1]
    )

    hist_y_pos = np.histogram(
        y, bins=40, density=True, range=[-1.5 * y_field / 2, 1.5 * y_field / 2]
    )
    x_hist_y_pos = hist_y_pos[1][:-1] + 0.5 * (hist_y_pos[1][1] - hist_y_pos[1][0])
    median_hist_y_pos = np.median(hist_y_pos[0][hist_y_pos[0] > 0])

    fwhm_y_pos = (
        -x_hist_y_pos[np.where(hist_y_pos[0] > median_hist_y_pos / 3)[0]][0]
        + x_hist_y_pos[np.where(hist_y_pos[0] > median_hist_y_pos / 3)[0]][-1]
    )
    fwhm_y_pos += x_hist_y_pos[2] - x_hist_y_pos[0]
    # a trick to correct the fact that the jaws aperture for alpha is a bit underestimated

    bool_x_pos = (fwhm_x_pos > x_field * (1 - tol)) & (fwhm_x_pos < x_field * (1 + tol))
    bool_y_pos = (fwhm_y_pos > y_field * (1 - tol)) & (fwhm_y_pos < y_field * (1 + tol))

    print("field dimension on the x-axis position :", x_field, "mm")
    print("FWHM measured for the x-axis beam :", fwhm_x_pos, "mm")
    print("field dimension on the y-axis position :", y_field, "mm")
    print("FWHM measured for the y-axis beam :", fwhm_y_pos, "mm")
    return bool_x_pos and bool_y_pos


if __name__ == "__main__":
    # paths
    paths = utility.get_default_test_paths(__file__, output_folder="test019_linac")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.g4_verbose = False
    # sim.visu = True
    sim.visu_type = "vrml"
    sim.check_volumes_overlap = False
    sim.number_of_threads = 1
    sim.output_dir = paths.output
    sim.random_seed = 123456789
    sim.check_volumes_overlap = True

    # unit
    nm = gate.g4_units.nm
    m = gate.g4_units.m
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV

    # world
    world = sim.world
    world.size = [2 * m, 2 * m, 2.1 * m]
    world.material = "G4_Galactic"
    a = np.array([0])

    # linac (empty)
    versa.add_linac_materials(sim)
    sad = 1000 * mm
    linac = versa.add_empty_linac_box(sim, "linac_box", sad)
    linac.material = "G4_Galactic"

    # jaws
    jaws = versa.add_jaws(sim, linac.name)

    # mlc
    mlc = versa.add_mlc(sim, linac.name)
    mlc_box = sim.volume_manager.get_volume(f"linac_box_mlc")
    mlc_box.material = "G4_Galactic"
    x_field = np.random.randint(10, 20, 1)[0] * cm
    y_field = np.random.randint(10, 20, 1)[0] * cm
    versa.set_rectangular_field(mlc, jaws, x_field, y_field, sad)

    # add alpha source
    plan = versa.add_phase_space_plane(sim, linac.name, 300)
    source = versa.add_phase_space_source(sim, plan.name)
    source.phsp_file = (
        paths.data / "output_ref" / "test019_linac" / "phsp_linac_mlc_alpha.root"
    )
    source.particle = "alpha"
    source.weight_key = None
    f = uproot.open(
        paths.data / "output_ref" / "test019_linac" / "phsp_linac_mlc_alpha.root"
    )
    data = f["linac_box_phsp_plane_phsp"].arrays()
    nb_part = len(data)

    source.n = nb_part / sim.number_of_threads
    if sim.visu:
        source.n = 20

    # physics
    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.set_production_cut("world", "all", 1000 * m)

    # add stat actor
    stats = sim.add_actor("SimulationStatisticsActor", "stats")
    stats.track_types_flag = True

    # add phase space
    plane = sim.add_volume("Box", "phsp_plane")
    plane.mother = world
    plane.material = "G4_Galactic"
    plane.size = [0.5 * m, 0.5 * m, 1 * nm]
    plane.color = [1, 0, 0, 1]  # red

    phsp = sim.add_actor("PhaseSpaceActor", f"phsp")
    phsp.attached_to = plane.name
    phsp.attributes = [
        "KineticEnergy",
        "Weight",
        "PrePosition",
        "PrePositionLocal",
        "PreDirection",
        "PreDirectionLocal",
        "PDGCode",
        "ParticleName",
    ]
    phsp.output_filename = "phsp_versa_mlc_wsrc.root"

    # start simulation
    sim.run()

    # print results
    print(stats)

    # end
    with uproot.open(phsp.get_output_path()) as f_phsp:
        arr = f_phsp["phsp"].arrays()
    is_ok = is_ok_test019(arr, x_field, y_field)
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
import test019_linac_phsp_helpers as test019

if __name__ == "__main__":
    # create sim
    sim = gate.Simulation()
    test019.create_simu_test019_phsp_source(sim, "_phsp_src_mt")

    # make it MT
    sim.number_of_threads = nt = 2

    sl = sim.get_source_user_info("phsp_source_local")
    sl.n /= nt
    sl.entry_start = [sl.n * p for p in range(nt)]
    sl.batch_size = sl.n

    sg = sim.get_source_user_info("phsp_source_global")
    sg.n /= nt
    sg.entry_start = [sl.n * p for p in range(nt)]
    sg.batch_size = sg.n

    print("source entry start", sl.entry_start)
    print("source entry start", sg.entry_start)

    # start simulation
    sim.run()

    # analyse
    is_ok = test019.analyse_test019_phsp_source(sim)

    utility.test_ok(is_ok)



from ..exception import fatal


def get_table_column(table, column_name):
    try:
        columns = table.cols
        try:
            return getattr(columns, column_name)
        except AttributeError:
            fatal(f"No column found for attribute {column_name}")
    except AttributeError:
        fatal(f"Incompatible input for 'table'. Should be a table. ")


def get_table_column_names(table):
    return table.cols._v_colnames


def get_node_name(node):
    return node._v_name


def set_node_name(node, name):
    node._v_name = name


def get_node_path(node, strip_leading_slash=False):
    p = node._v_pathname
    if strip_leading_slash is True:
        p = p.lstrip("/")
    return p


def get_group_name(group):
    return get_node_name(group)


def get_nodes_in_group(group, node_type=None, node_name=None, condition_functions=None):
    all_nodes = [n for n in group._f_iter_nodes()]
    return filter_nodes(all_nodes, node_type, node_name, condition_functions)


def filter_nodes(nodes, node_type=None, node_name=None, condition_functions=None):
    conditions = []
    if node_type is not None:
        conditions.append(lambda _n: isinstance(_n, node_type))
    if node_name is not None:
        conditions.append(lambda _n: get_node_name(_n) == node_name)
    if condition_functions is not None:
        conditions.extend(tuple(condition_functions))
    selected_nodes = []
    for n in nodes:
        if all([c(n) is True for c in conditions]):
            selected_nodes.append(n)
    return selected_nodes


def get_parent(node):
    return node._v_parent


def create_hard_links_to_nodes_in_group(where_group, target_group):
    file_handle = where_group._v_file
    for n in target_group._f_iter_nodes():
        print(f"where_group: {where_group}")
        print(f"node name: {get_node_name(n)}")
        print(f"n: {n}")
        file_handle.create_hard_link(where_group, get_node_name(n), target=n)


def get_create_function(file_handle, which):
    if which in ("table", "Table"):
        create_func = file_handle.create_table
    elif which in ("array", "Array"):
        create_func = file_handle.create_array
    elif which in ("carray", "Carray"):
        create_func = file_handle.create_carray
    elif which in ("earray", "Earray"):
        create_func = file_handle.create_earray
    elif which in ("group", "Group"):
        create_func = file_handle.create_group
    else:
        fatal(
            f"Unkown kind of output structure '{which}'. "
            f"Known structures are table, array, carray, earray, group"
        )
    return create_func



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from scipy.spatial.transform import Rotation
import numpy as np
from anytree import Node, RenderTree
import uproot


def test082_test(df):
    df = df[df["PDGCode"] == 22]
    nb_event = len(df["ParentID"])
    nb_event_to_interest = len(df["ParentID"][df["ParentID"] == 0])

    tab_vertex_ekin = df["TrackVertexKineticEnergy"]
    tab_ekin = df["KineticEnergy"]
    dz_diff = df["PreDirection_Z"][
        (df["PreDirection_Z"] != -1)
        & (df["TrackVertexKineticEnergy"] == df["KineticEnergy"])
    ]

    print(
        "Number of detected photons with an energy/a modified momentum:", len(dz_diff)
    )
    if (nb_event_to_interest == nb_event) and (
        np.all(tab_ekin == tab_vertex_ekin) and (len(dz_diff) == 0)
    ):
        return True
    return False


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__)
    output_path = paths.output

    print(output_path)
    # create the simulation
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    # ui.visu = True
    ui.visu_type = "vrml"
    ui.check_volumes_overlap = False
    # ui.running_verbose_level = gate.logger.EVENT
    ui.number_of_threads = 1
    ui.random_seed = "auto"

    # units
    m = gate.g4_units.m
    km = gate.g4_units.km
    mm = gate.g4_units.mm
    cm = gate.g4_units.cm
    nm = gate.g4_units.nm
    Bq = gate.g4_units.Bq
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    sec = gate.g4_units.s
    gcm3 = gate.g4_units["g/cm3"]

    sim.volume_manager.material_database.add_material_weights(
        "Tungsten",
        ["W"],
        [1],
        19.3 * gcm3,
    )

    #  adapt world size
    world = sim.world
    world.size = [0.2 * m, 0.2 * m, 0.2 * m]
    world.material = "G4_Galactic"

    source = sim.add_source("GenericSource", "photon_source")
    source.particle = "gamma"
    source.position.type = "box"
    source.attached_to = world.name
    source.position.size = [1 * nm, 1 * nm, 1 * nm]
    source.position.translation = [0, 0, 10 * cm + 1 * mm]
    source.direction.type = "momentum"
    source.direction_relative_to_attached_volume = True
    # source1.direction.focus_point = [0*cm, 0*cm, -5 *cm]
    source.direction.momentum = [0, 0, -1]
    source.energy.type = "mono"
    source.energy.mono = 1 * MeV
    source.n = 100000

    tungsten = sim.add_volume("Box", "tungsten_box")
    tungsten.size = [3 * cm, 3 * cm, 1 * cm]
    tungsten.material = "Tungsten"
    tungsten.mother = world.name
    tungsten.color = [0.5, 0.9, 0.3, 1]

    kill_proc_act = sim.add_actor("KillAccordingProcessesActor", "kill_proc_act")
    kill_proc_act.attached_to = tungsten.name
    kill_proc_act.processes_to_kill = ["compt", "Rayl"]

    phsp_sphere = sim.add_volume("Sphere", "phsp_sphere")
    phsp_sphere.mother = world.name
    phsp_sphere.material = "G4_Galactic"
    phsp_sphere.rmin = 5 * cm
    phsp_sphere.rmax = 5 * cm + 1 * nm

    sim.output_dir = output_path
    phsp = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    phsp.attached_to = phsp_sphere.name
    phsp.attributes = [
        "ParentID",
        "EventID",
        "TrackID",
        "KineticEnergy",
        "TrackVertexKineticEnergy",
        "PreDirection",
        "PDGCode",
    ]
    name_phsp = "test082_" + phsp.name + ".root"
    phsp.output_filename = name_phsp

    sim.physics_manager.physics_list_name = "G4EmStandardPhysics_option3"
    sim.physics_manager.enable_decay = False
    sim.physics_manager.global_production_cuts.gamma = 1 * km
    sim.physics_manager.global_production_cuts.electron = 1 * km
    sim.physics_manager.global_production_cuts.positron = 1 * km

    # Mandatory for this actor, since gamma processes are encompassed in GammaGeneralProc without.
    s = f"/process/em/UseGeneralProcess false"
    sim.g4_commands_before_init.append(s)

    #
    # s = sim.add_actor("SimulationStatisticsActor", "Stats")
    # s.track_types_flag = True
    #
    # # go !
    sim.run()
    #
    phsp = uproot.open(str(output_path) + "/test082_PhaseSpace.root" + ":PhaseSpace")

    df = phsp.arrays()
    is_ok = test082_test(df)
    #
    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from opengate.contrib.spect.spect_config import *
from test090_helpers import *

if __name__ == "__main__":
    paths = utility.get_default_test_paths(
        __file__, None, output_folder="test090_spect_config_3"
    )

    # TEST 3: intevo, Lu177, 2 heads, 3 angles, with FF-AA, primary only
    sc = create_test_spect_config(paths)

    # units
    deg = g4_units.deg
    Bq = g4_units.Bq
    cm = g4_units.cm

    # make it FF-AA primary
    sc.free_flight_config.primary_activity = 1e5 * Bq
    sc.free_flight_config.max_rejection = 10000
    sc.free_flight_config.angle_tolerance = 15 * deg
    sc.free_flight_config.forced_direction_flag = False

    # create the simulation
    sim = gate.Simulation()
    sim.random_seed = 654456
    sc.setup_simulation_ff_primary(sim, visu=False)
    stats = sim.actor_manager.find_actors("stats")[0]

    # go
    sim.run(start_new_process=True)

    # we check only that the output files exist
    is_ok = True
    is_ok = check_stats_file(18960, sc, stats, is_ok)
    is_ok = check_projection_files(sim, paths, stats, is_ok, tol=60, squared_flag=True)

    utility.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from opengate.tests import utility
from opengate.sources.utility import get_spectrum


if __name__ == "__main__":
    paths = utility.get_default_test_paths(__file__, "", output_folder="test084")

    # create the simulation
    sim = gate.Simulation()

    # main options
    sim.output_dir = paths.output

    # units
    m = gate.g4_units.m
    cm = gate.g4_units.cm
    MeV = gate.g4_units.MeV
    keV = gate.g4_units.keV
    Bq = gate.g4_units.Bq
    mm = gate.g4_units.mm
    gcm3 = gate.g4_units.g_cm3

    #  change world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # add a material database
    sim.volume_manager.add_material_database(paths.data / "GateMaterials.db")

    # image
    patient = sim.add_volume("Image", "patient")
    patient.image = paths.data / "patient-4mm.mhd"
    patient.material = "G4_AIR"  # material used by default
    f1 = str(paths.data / "Schneider2000MaterialsTable.txt")
    f2 = str(paths.data / "Schneider2000DensitiesTable.txt")
    tol = 0.01 * gcm3
    print(f"HU density tolerance: {tol/gcm3} gcm3")
    patient.voxel_materials, materials = (
        gate.geometry.materials.HounsfieldUnit_to_material(sim, tol, f1, f2)
    )
    print(f"Number of materials: {len(materials)}")

    # mu map actor (process at the first "begin of run" only)
    mumap = sim.add_actor("AttenuationImageActor", "mumap")
    mumap.image_volume = patient  # FIXME volume for the moment, not the name
    mumap.output_filename = "mumap2.mhd"
    energies = get_spectrum("Lu177", "gamma", "radar").energies
    mumap.energy = energies[3]  # 0.208 keV
    mumap.database = "EPDL"  # NIST or EPDL
    print(f"Energy is {mumap.energy/keV} keV")
    print(f"Database is {mumap.database}")

    # remove verbose
    sim.verbose_level = "NONE"
    sim.run(start_new_process=True)

    # compare with ref
    ref = paths.output_ref / mumap.output_filename
    is_ok = utility.assert_images(
        ref,
        mumap.get_output_path(),
        tolerance=1e-5,
        fig_name=paths.output / "mumap2.png",
        sum_tolerance=1e-5,
    )

    utility.test_ok(is_ok)
